import os
import json
import torch

import numpy as np

from typing import *
from cftool.dist import Parallel
from cftool.misc import shallow_copy_dict
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.misc import LoggingMixin

from .task import Task
from ..types import data_type
from ..pipeline import Pipeline
from .runs._utils import meta_config_name
from .runs._utils import data_config_file


def _task(
    task: Task,
    execute: str,
    config_folder: str,
    cuda: Optional[Union[int, str]] = None,
) -> None:
    task.run(execute, config_folder, cuda)


def inject_distributed_tqdm_kwargs(
    i: int,
    num_jobs: int,
    kwargs: Dict[str, Any],
) -> None:
    kwargs.setdefault("use_tqdm", True)
    kwargs.setdefault("use_step_tqdm", False)
    kwargs.setdefault("in_distributed", True)
    kwargs.setdefault("tqdm_position", i % (num_jobs or 1))
    kwargs.setdefault("tqdm_desc", f"epoch (task {i})")


class ExperimentResults(NamedTuple):
    workplaces: List[str]
    workplace_keys: List[Tuple[str, str]]
    pipelines: Optional[List[Pipeline]]

    @property
    def pipeline_dict(self) -> Dict[str, Pipeline]:
        if self.pipelines is None:
            raise ValueError("pipelines are not provided")
        return dict(zip(self.workplaces, self.pipelines))


class Experiment(LoggingMixin):
    tasks_folder = "__tasks__"
    default_root_workplace = "__experiment__"

    def __init__(
        self,
        *,
        num_jobs: int = 1,
        use_cuda: bool = True,
        available_cuda_list: Optional[List[int]] = None,
        resource_config: Optional[Dict[str, Any]] = None,
    ):
        use_cuda = use_cuda and torch.cuda.is_available()
        if available_cuda_list is None and not use_cuda:
            available_cuda_list = []
        self.num_jobs = num_jobs
        self.use_cuda = use_cuda
        self.cuda_list = available_cuda_list
        self.resource_config = resource_config or {}
        self.tasks: Dict[Tuple[str, str], Task] = {}
        self.executes: Dict[Tuple[str, str], str] = {}
        self.workplaces: Dict[Tuple[str, str], str] = {}
        self.results: Optional[ExperimentResults] = None

    @staticmethod
    def data_folder(workplace: Optional[str] = None) -> str:
        workplace = workplace or Experiment.default_root_workplace
        return os.path.join(workplace, "__data__")

    @staticmethod
    def dump_data(
        data_folder: str,
        x: data_type,
        y: data_type = None,
        postfix: str = "",
    ) -> None:
        for key, value in zip([f"x{postfix}", f"y{postfix}"], [x, y]):
            if value is None:
                continue
            np.save(os.path.join(data_folder, f"{key}.npy"), value)

    @staticmethod
    def dump_data_bundle(
        x: data_type,
        y: data_type = None,
        x_cv: data_type = None,
        y_cv: data_type = None,
        *,
        workplace: Optional[str] = None,
        data_folder: Optional[str] = None,
    ) -> str:
        if data_folder is None:
            data_folder = Experiment.data_folder(workplace)
        os.makedirs(data_folder, exist_ok=True)
        if not isinstance(x, np.ndarray):
            data_config = {"x": x, "y": y, "x_cv": x_cv, "y_cv": y_cv}
            for k, v in data_config.items():
                if v is not None:
                    assert isinstance(v, str)
                    data_config[k] = os.path.abspath(v)
            with open(os.path.join(data_folder, data_config_file), "w") as f:
                json.dump(data_config, f)
        else:
            Experiment.dump_data(data_folder, x, y)
            Experiment.dump_data(data_folder, x_cv, y_cv, "_cv")
        return data_folder

    @staticmethod
    def fetch_data(
        postfix: str = "",
        *,
        workplace: Optional[str] = None,
        data_folder: Optional[str] = None,
    ) -> Tuple[data_type, data_type]:
        if data_folder is None:
            data_folder = Experiment.data_folder(workplace)
        data = []
        for key in [f"x{postfix}", f"y{postfix}"]:
            file = os.path.join(data_folder, f"{key}.npy")
            data.append(None if not os.path.isfile(file) else np.load(file))
        return data[0], data[1]

    @staticmethod
    def workplace(workplace_key: Tuple[str, str], root_workplace: str) -> str:
        workplace = os.path.join(root_workplace, *workplace_key)
        return os.path.abspath(workplace)

    @property
    def num_tasks(self) -> int:
        return len(self.tasks)

    def add_task(
        self,
        x: data_type = None,
        y: data_type = None,
        x_cv: data_type = None,
        y_cv: data_type = None,
        *,
        model: str = "fcnn",
        execute: str = "basic",
        root_workplace: Optional[str] = None,
        workplace_key: Optional[Tuple[str, str]] = None,
        config: Optional[Dict[str, Any]] = None,
        data_folder: Optional[str] = None,
        **task_meta_kwargs: Any,
    ) -> str:
        if workplace_key is None:
            counter = 0
            while True:
                workplace_key = model, str(counter)
                if workplace_key not in self.tasks:
                    break
                counter += 1
        if workplace_key in self.tasks:
            raise ValueError(f"task already exists with '{workplace_key}'")
        root_workplace = root_workplace or self.default_root_workplace
        workplace = self.workplace(workplace_key, root_workplace)
        copied_config = shallow_copy_dict(config or {})
        copied_config["model"] = model
        inject_distributed_tqdm_kwargs(len(self.tasks), self.num_jobs, copied_config)
        increment_config = {}
        if data_folder is None and x is not None:
            data_folder = self.dump_data_bundle(x, y, x_cv, y_cv, workplace=workplace)
        if data_folder is not None:
            increment_config["data_folder"] = os.path.abspath(data_folder)
        increment_config["verbose_level"] = copied_config.get("verbose_level", 0)
        increment_config["trigger_logging"] = copied_config.get("trigger_logging", True)
        new_task = Task(
            workplace=workplace,
            config=copied_config,
            increment_config=increment_config,
            **shallow_copy_dict(task_meta_kwargs),
        )
        self.tasks[workplace_key] = new_task
        self.executes[workplace_key] = execute
        self.workplaces[workplace_key] = workplace
        return workplace

    def run_tasks(
        self,
        *,
        use_tqdm: bool = True,
        task_loader: Optional[Callable[[str], Pipeline]] = None,
    ) -> ExperimentResults:
        resource_config = shallow_copy_dict(self.resource_config)
        gpu_config = resource_config.setdefault("gpu_config", {})
        gpu_config["available_cuda_list"] = self.cuda_list
        parallel = Parallel(
            self.num_jobs,
            use_tqdm=use_tqdm,
            use_cuda=self.use_cuda,
            resource_config=resource_config,
        )
        sorted_workplace_keys = sorted(self.tasks)
        sorted_tasks = [self.tasks[key] for key in sorted_workplace_keys]
        sorted_executes = [self.executes[key] for key in sorted_workplace_keys]
        sorted_workplaces = [self.workplaces[key] for key in sorted_workplace_keys]
        parallel(_task, sorted_tasks, sorted_executes, sorted_workplaces)
        if task_loader is None:
            pipelines = None
        else:
            pipelines = list(map(task_loader, sorted_workplaces))
        self.results = ExperimentResults(
            sorted_workplaces,
            sorted_workplace_keys,
            pipelines,
        )
        return self.results

    def save(self, export_folder: str, *, compress: bool = True) -> "Experiment":
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            Saving.prepare_folder(self, export_folder)
            # tasks
            tasks_folder = os.path.join(abs_folder, self.tasks_folder)
            for workplace_key, task in self.tasks.items():
                task.save(os.path.join(tasks_folder, *workplace_key))
            # meta
            if self.results is None:
                raise ValueError("results are not generated yet")
            meta_config = {
                "executes": self.executes,
                "num_jobs": self.num_jobs,
                "use_cuda": self.use_cuda,
                "cuda_list": self.cuda_list,
                "resource_config": self.resource_config,
                "results": ExperimentResults(
                    self.results.workplaces,
                    self.results.workplace_keys,
                    None,
                ),
            }
            Saving.save_dict(meta_config, meta_config_name, abs_folder)
            if compress:
                Saving.compress(abs_folder, remove_original=True)
        return self

    @classmethod
    def load(
        cls,
        saving_folder: str,
        *,
        compress: bool = True,
        task_loader: Optional[Callable[[str], Pipeline]] = None,
    ) -> "Experiment":
        workplace: str
        workplace_key: Tuple[str, str]

        abs_folder = os.path.abspath(saving_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [saving_folder]):
            with Saving.compress_loader(abs_folder, compress, remove_extracted=False):
                meta_config = Saving.load_dict(meta_config_name, abs_folder)
                experiment = cls(
                    num_jobs=meta_config["num_jobs"],
                    use_cuda=meta_config["use_cuda"],
                    available_cuda_list=meta_config["cuda_list"],
                    resource_config=meta_config["resource_config"],
                )
                experiment.executes = meta_config["executes"]
                results = list(meta_config["results"])
                # tasks
                pipelines = []
                experiment.tasks = {}
                tasks_folder = os.path.join(abs_folder, cls.tasks_folder)
                for workplace, workplace_key in zip(*results[:2]):
                    task_folder = os.path.join(tasks_folder, *workplace_key)
                    experiment.tasks[workplace_key] = Task.load(task_folder)
                    if task_loader is not None:
                        pipelines.append(task_loader(workplace))
                results[-1] = pipelines or None
                experiment.results = ExperimentResults(*results)
        return experiment


__all__ = [
    "inject_distributed_tqdm_kwargs",
    "Experiment",
]



import os
import math
import torch

import numpy as np

from typing import *
from tqdm.autonotebook import tqdm
from cftool.misc import update_dict
from cftool.misc import shallow_copy_dict
from cfdata.tabular import task_type_type
from cftool.ml.utils import collate_fn_type
from cftool.ml.utils import Metrics
from torch.nn.functional import one_hot

from .basic import *
from ..misc.toolkit import *
from .register import register_metric
from ..types import data_type
from ..pipeline import Pipeline
from ..protocol import DataProtocol


class EnsembleResults(NamedTuple):
    data: DataProtocol
    pipelines: List[Pipeline]
    pattern_weights: Optional[np.ndarray]
    predict_config: Optional[Dict[str, Any]]

    @property
    def pattern(self) -> EnsemblePattern:
        predict_config = self.predict_config or {}
        patterns = [m.to_pattern(**predict_config) for m in self.pipelines]
        return Ensemble.stacking(patterns, pattern_weights=self.pattern_weights)


class MetricsPlaceholder(NamedTuple):
    config: Dict[str, Any]


class Ensemble:
    def __init__(
        self,
        task_type: task_type_type,
        config: Optional[Dict[str, Any]] = None,
    ):
        self.task_type = task_type
        if config is None:
            config = {}
        self.config = shallow_copy_dict(config)

    @staticmethod
    def stacking(
        patterns: List[ModelPattern],
        *,
        pattern_weights: Optional[np.ndarray] = None,
        ensemble_method: Optional[Union[str, collate_fn_type]] = None,
    ) -> EnsemblePattern:
        if ensemble_method is None:
            if pattern_weights is None:
                ensemble_method = "default"
            else:
                if abs(pattern_weights.sum() - 1.0) > 1e-4:
                    raise ValueError("`pattern_weights` should sum to 1.0")
                pattern_weights = pattern_weights.reshape([-1, 1, 1])

                def ensemble_method(
                    arrays: List[np.ndarray],
                    requires_prob: bool,
                ) -> np.ndarray:
                    shape = [len(arrays), len(arrays[0]), -1]
                    predictions = np.array(arrays).reshape(shape)
                    if requires_prob or not is_int(predictions):
                        return (predictions * pattern_weights).sum(axis=0)
                    encodings = one_hot(to_torch(predictions).to(torch.long).squeeze())
                    encodings = encodings.to(torch.float32)
                    weighted = (encodings * pattern_weights).sum(dim=0)
                    return to_numpy(weighted.argmax(1)).reshape([-1, 1])

        return EnsemblePattern(patterns, ensemble_method)

    def bagging(
        self,
        x: data_type,
        y: data_type = None,
        *,
        k: int = 10,
        num_jobs: int = 1,
        model: str = "fcnn",
        model_configs: Optional[Dict[str, Dict[str, Any]]] = None,
        predict_config: Optional[Dict[str, Any]] = None,
        sequential: Optional[bool] = None,
        temp_folder: str = "__tmp__",
        return_patterns: bool = True,
        use_tqdm: bool = True,
    ) -> EnsembleResults:
        repeat_result = repeat_with(
            x,
            y,
            models=model,
            model_configs=model_configs,
            predict_config=predict_config,
            sequential=sequential,
            num_jobs=num_jobs,
            num_repeat=k,
            temp_folder=temp_folder,
            return_patterns=return_patterns,
            use_tqdm=use_tqdm,
            **self.config,
        )

        data = repeat_result.data
        pipelines = repeat_result.pipelines
        assert data is not None and pipelines is not None
        return EnsembleResults(data, pipelines[model], None, predict_config)

    def adaboost(
        self,
        x: data_type,
        y: data_type = None,
        *,
        k: int = 10,
        eps: float = 1e-12,
        model: str = "fcnn",
        temp_folder: str = "__tmp__",
        predict_config: Optional[Dict[str, Any]] = None,
        increment_config: Optional[Dict[str, Any]] = None,
        sample_weights: Optional[np.ndarray] = None,
    ) -> EnsembleResults:
        if increment_config is None:
            increment_config = {}
        config = shallow_copy_dict(self.config)
        update_dict(increment_config, config)
        config["cv_split"] = 0.0
        config.setdefault("use_tqdm", False)
        config.setdefault("use_binary_threshold", False)
        config.setdefault("verbose_level", 0)

        @register_metric("adaboost_error", -1, False)
        def adaboost_error(
            self_: Union[Metrics, MetricsPlaceholder],
            target_: np.ndarray,
            predictions_: np.ndarray,
        ) -> float:
            target_ = target_.astype(np.float32)
            predictions_ = predictions_.astype(np.float32)
            sample_weights_ = self_.config.get("sample_weights")
            errors = (target_ != predictions_).ravel()
            if sample_weights_ is None:
                e_ = errors.mean()
            else:
                e_ = sample_weights_[errors].sum() / len(errors)
            return e_.item()

        data = None
        pipelines = []
        patterns, pattern_weights = [], []
        for i in tqdm(list(range(k))):
            cfg = shallow_copy_dict(config)
            cfg["logging_folder"] = os.path.join(temp_folder, str(i))
            metric_config = {"sample_weights": sample_weights}
            if sample_weights is not None:
                cfg["metric_config"] = {
                    "types": "adaboost_error",
                    "adaboost_error_config": metric_config,
                }
            m = make(model=model, **cfg)
            m.fit(x, y, sample_weights=sample_weights)
            metrics_placeholder = MetricsPlaceholder(metric_config)
            predictions: np.ndarray = m.predict(x, contains_labels=True)
            predictions = predictions.astype(np.float32)
            target = m.data.processed.y.astype(np.float32)
            e = adaboost_error(metrics_placeholder, target, predictions)
            em = min(max(e, eps), 1.0 - eps)
            am = 0.5 * math.log(1.0 / em - 1.0)
            if sample_weights is None:
                sample_weights = np.ones_like(predictions).ravel()
            target[target == 0.0] = predictions[predictions == 0.0] = -1.0
            sample_weights *= np.exp(-am * target * predictions).ravel()
            sample_weights /= np.mean(sample_weights)
            patterns.append(m.to_pattern())
            pattern_weights.append(am)
            if data is None:
                data = m.data
            pipelines.append(m)

        weights_array = np.array(pattern_weights, np.float32)
        weights_array /= weights_array.sum()

        assert data is not None
        return EnsembleResults(data, pipelines, weights_array, predict_config)


__all__ = [
    "Ensemble",
    "EnsembleResults",
]



import cflearn

from cflearn.api.hpo import OptunaArgs

from ._utils import get_info


if __name__ == "__main__":
    info = get_info(requires_data=False)
    cflearn.optuna_core(
        OptunaArgs(
            info.meta["cuda"],
            info.meta["compress"],
            info.meta["num_trial"],
            info.meta["task_config_folder"],
            info.meta["key_mapping_folder"],
        )
    )



from typing import Any
from typing import Dict

from ..base import HeadConfigs


@HeadConfigs.register("linear", "default")
class DefaultLinearConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        return {"linear_config": {}}


__all__ = ["DefaultLinearConfig"]



import torch

from ..base import ExtractorBase


@ExtractorBase.register("identity")
class Identity(ExtractorBase):
    @property
    def out_dim(self) -> int:
        return self.in_flat_dim

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        return net


@ExtractorBase.register("identity_ts")
class IdentityTS(ExtractorBase):
    @property
    def out_dim(self) -> int:
        return self.in_dim

    @property
    def flatten_ts(self) -> bool:
        return False

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        return net


__all__ = ["Identity", "IdentityTS"]



import numpy as np
import torch.nn as nn

from .base import ModelBase
from ..misc.toolkit import to_numpy
from ..modules.blocks import Linear


@ModelBase.register("ndt")
@ModelBase.register_pipe("ndt")
class NDT(ModelBase):
    @property
    def head(self) -> nn.Module:
        return self.heads["ndt"]

    @property
    def to_planes(self) -> Linear:
        return self.head.to_planes  # type: ignore

    @property
    def to_routes(self) -> Linear:
        return self.head.to_routes  # type: ignore

    @property
    def to_leaves(self) -> Linear:
        return self.head.to_leaves  # type: ignore

    @property
    def hyperplane_weights(self) -> np.ndarray:
        return to_numpy(self.to_planes.linear.weight)

    @property
    def hyperplane_thresholds(self) -> np.ndarray:
        return to_numpy(-self.to_planes.linear.bias)

    @property
    def route_weights(self) -> np.ndarray:
        return to_numpy(self.to_routes.linear.weight)

    @property
    def class_log_distributions(self) -> np.ndarray:
        return to_numpy(self.to_leaves.linear.weight)

    @property
    def class_log_prior(self) -> np.ndarray:
        return to_numpy(self.to_leaves.linear.bias)

    @property
    def class_prior(self) -> np.ndarray:
        return np.exp(self.class_log_prior)


__all__ = ["NDT"]



from typing import Any
from typing import Dict

from ..base import HeadConfigs


@HeadConfigs.register("dndf", "default")
class DefaultDNDFConfig(HeadConfigs):
    def get_default(self) -> Dict[str, Any]:
        return {
            "dndf_config": {
                "is_regression": self.tr_data.is_reg,
                "tree_proj_config": None,
            }
        }


__all__ = [
    "DefaultDNDFConfig",
]



import torch

from torch import Tensor
from typing import Any
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from torch.nn import Module
from torch.nn import ModuleList

from ...blocks import MLP
from ...blocks import CrossBase
from ...blocks import CrossBlock
from ...blocks import ConditionalBlocks
from ...blocks import MonotonousMapping
from ....types import tensor_tuple_type


def transition_builder(dim: int) -> Module:
    h_dim = int(dim // 2)
    return MonotonousMapping.make_couple(h_dim, h_dim, h_dim, "sigmoid", ascent=True)


def get_cond_mappings(
    condition_dim: int,
    cond_out_dim: Optional[int],
    num_units: List[int],
    to_latent: bool,
) -> ModuleList:
    cond_module = MLP.simple(
        condition_dim,
        cond_out_dim,
        num_units,
        bias=not to_latent,
        activation="glu",
    )
    return cond_module.mappings


responses_tuple_type = Tuple[List[Tensor], List[Tensor]]


class Pack(NamedTuple):
    net: Tensor
    cond: Tensor
    responses_tuple: responses_tuple_type


class MonoSplit(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        ascent1: bool,
        ascent2: bool,
        to_latent: bool,
        builder: Callable[[int, int, bool], ConditionalBlocks],
    ) -> None:
        super().__init__()
        self.to_latent = to_latent
        self.m1 = builder(in_dim, out_dim, ascent1)
        self.m2 = builder(in_dim, out_dim, ascent2)

    def forward(
        self,
        net: Union[Tensor, tensor_tuple_type],
        cond: Union[Tensor, responses_tuple_type],
    ) -> Union[tensor_tuple_type, Pack]:
        if self.to_latent:
            return self.m1(net, cond), self.m2(net, cond)
        cond1: Union[Tensor, List[Tensor]]
        cond2: Union[Tensor, List[Tensor]]
        if isinstance(cond, tuple):
            cond1, cond2 = cond
        else:
            cond1 = cond2 = cond
        o1, o2 = self.m1(net[0], cond1), self.m2(net[1], cond2)
        merged_net = 0.5 * (o1.net + o2.net)
        merged_cond = 0.5 * (o1.cond + o2.cond)
        return Pack(merged_net, merged_cond, (o1.responses, o2.responses))


class MonoCross(CrossBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        centralize: bool = True,
        **kwargs: Any,
    ):
        super().__init__()
        kwargs["bias"] = not centralize
        self.mapping = MonotonousMapping(in_dim, out_dim, ascent=True, **kwargs)
        self.centralize = centralize

    def forward(self, net: Tensor, cond: Tensor) -> Tensor:
        crossed = self.mapping(net * cond.abs()) * torch.sigmoid(cond)
        if self.centralize:
            return crossed
        return crossed + cond

    def extra_repr(self) -> str:
        return f"(centralize): {self.centralize}"

    @classmethod
    def make(
        cls,
        dim: int,
        inner: bool,
        *,
        centralize: bool = True,
        **kwargs: Any,
    ) -> "MonoCross":
        out_dim = 1 if inner else dim
        return cls(dim, out_dim, centralize=centralize, **kwargs)


def get_q_cross_builder(to_latent: bool) -> Callable[[int], Module]:
    return lambda dim: MonoCross.make(dim, to_latent, centralize=to_latent)


def get_y_cross_builder(to_latent: bool) -> Callable[[int], Module]:
    return lambda dim: MonoCross.make(dim, to_latent, centralize=not to_latent)


def monotonous_builder(
    is_q: bool,
    ascent1: bool,
    ascent2: bool,
    to_latent: bool,
    num_layers: int,
    condition_dim: int,
    cond_mappings: Optional[ModuleList] = None,
) -> Callable[[int, int], Module]:
    def _core(
        in_dim: int,
        out_dim: int,
        ascent: bool,
    ) -> ConditionalBlocks:
        cond_out_dim: Optional[int]
        block_out_dim: Optional[int]
        if to_latent:
            num_units = [out_dim] * (num_layers + 1)
            cond_out_dim = block_out_dim = None
        else:
            num_units = [in_dim] * num_layers
            # only cdf will initialize cond mappings here
            # q will utilize median mappings defined outside
            cond_out_dim = 1
            # block : y_mul / cdf_mul
            block_out_dim = 1

        blocks = MonotonousMapping.stack(
            in_dim,
            block_out_dim,
            num_units,
            ascent=ascent,
            use_couple_bias=(is_q and not to_latent) or (not is_q and to_latent),
            activation="sigmoid" if to_latent else "tanh",
            return_blocks=True,
        )
        assert isinstance(blocks, list)

        if cond_mappings is not None:
            cond_mappings_ = cond_mappings
        else:
            cond_mappings_ = get_cond_mappings(
                condition_dim,
                cond_out_dim,
                num_units,
                to_latent,
            )

        if is_q:
            cross_builder = get_q_cross_builder(to_latent)
        else:
            cross_builder = get_y_cross_builder(to_latent)
        cond_mixtures = ModuleList(
            [
                CrossBlock(
                    unit,
                    False,
                    cross_builder=cross_builder,
                )
                for unit in num_units
            ]
        )
        return ConditionalBlocks(
            ModuleList(blocks),
            cond_mappings_,
            add_last=to_latent,
            detach_condition=not to_latent,
            cond_mixtures=cond_mixtures,
        )

    def _split_core(in_dim: int, out_dim: int) -> Module:
        if not to_latent:
            in_dim = int(in_dim // 2)
        else:
            out_dim = int(out_dim // 2)
        return MonoSplit(in_dim, out_dim, ascent1, ascent2, to_latent, _core)

    return _split_core


__all__ = [
    "get_cond_mappings",
    "monotonous_builder",
    "transition_builder",
    "responses_tuple_type",
    "Pack",
]



from typing import Any
from typing import Dict

from ....configs import Configs


@Configs.register("transformer", "default")
class DefaultTransformerConfig(Configs):
    def get_default(self) -> Dict[str, Any]:
        return {
            "num_heads": 8,
            "num_layers": 6,
            "latent_dim": 256,
            "norm": None,
            "input_linear_config": {"bias": False},
            "transformer_layer_config": {},
        }


__all__ = ["DefaultTransformerConfig"]



import os
import json
import math
import torch
import pprint
import logging

import numpy as np
import torch.nn as nn

from abc import abstractmethod
from abc import ABC
from abc import ABCMeta
from typing import Any
from typing import Set
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from functools import partial
from tqdm.autonotebook import tqdm
from cftool.ml import Metrics
from cftool.misc import register_core
from cftool.misc import timing_context
from cftool.misc import shallow_copy_dict
from cftool.misc import context_error_handler
from cfdata.types import np_int_type
from cfdata.tabular import data_type
from cfdata.tabular import str_data_type
from cfdata.tabular import TaskTypes
from cfdata.tabular import DataTuple
from cfdata.tabular.recognizer import Recognizer

from .types import np_dict_type
from .types import tensor_dict_type
from .types import loader_batch_type
from .types import prefetch_batch_type
from .misc.toolkit import to_prob
from .misc.toolkit import is_float
from .misc.toolkit import to_numpy
from .misc.toolkit import to_torch
from .misc.toolkit import eval_context
from .misc.toolkit import LoggingMixinWithRank
from .modules.blocks import EMA

data_dict: Dict[str, Type["DataProtocol"]] = {}
sampler_dict: Dict[str, Type["SamplerProtocol"]] = {}
loader_dict: Dict[str, Type["DataLoaderProtocol"]] = {}


class DataSplit(NamedTuple):
    split: "DataProtocol"
    remained: "DataProtocol"
    split_indices: np.ndarray
    remained_indices: np.ndarray


class DataProtocol(ABC):
    is_ts: bool
    is_clf: bool
    is_simplify: bool
    raw_dim: int
    num_classes: int

    raw: DataTuple
    converted: DataTuple
    processed: DataTuple
    ts_indices: Set[int]
    recognizers: Dict[int, Optional[Recognizer]]

    _verbose_level: int
    _has_column_names: bool

    @abstractmethod
    def __init__(self, **kwargs: Any):
        self._verbose_level = kwargs.get("verbose_level", 2)

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def read_file(
        self,
        file_path: str,
        *,
        contains_labels: bool = True,
    ) -> Tuple[str_data_type, Optional[str_data_type]]:
        pass

    @abstractmethod
    def read(
        self,
        x: Union[str, data_type],
        y: Optional[Union[int, data_type]] = None,
        *,
        contains_labels: bool = True,
        **kwargs: Any,
    ) -> "DataProtocol":
        pass

    @abstractmethod
    def split(self, n: Union[int, float], *, order: str = "auto") -> DataSplit:
        pass

    @abstractmethod
    def split_with_indices(
        self,
        split_indices: np.ndarray,
        remained_indices: np.ndarray,
    ) -> DataSplit:
        pass

    @abstractmethod
    def transform(
        self,
        x: Union[str, data_type],
        y: data_type = None,
        **kwargs: Any,
    ) -> DataTuple:
        pass

    @abstractmethod
    def transform_labels(self, y: data_type) -> np.ndarray:
        pass

    @abstractmethod
    def recover_labels(self, y: np.ndarray, *, inplace: bool = False) -> np.ndarray:
        pass

    @abstractmethod
    def copy_to(
        self,
        x: Union[str, data_type],
        y: data_type = None,
        *,
        contains_labels: bool = True,
    ) -> "DataProtocol":
        pass

    @abstractmethod
    def save(
        self,
        folder: str,
        *,
        compress: bool = True,
        retain_data: bool = True,
        remove_original: bool = True,
    ) -> "DataProtocol":
        pass

    @classmethod
    @abstractmethod
    def load(
        cls,
        folder: str,
        *,
        compress: bool = True,
        verbose_level: int = 0,
    ) -> "DataProtocol":
        pass

    @property
    def is_reg(self) -> bool:
        return not self.is_clf

    @property
    def task_type(self) -> TaskTypes:
        if not self.is_ts:
            if self.is_clf:
                return TaskTypes.CLASSIFICATION
            return TaskTypes.REGRESSION
        if self.is_clf:
            return TaskTypes.TIME_SERIES_CLF
        return TaskTypes.TIME_SERIES_REG

    @classmethod
    def get(cls, name: str) -> Type["DataProtocol"]:
        return data_dict[name]

    @classmethod
    def make(cls, name: str, **kwargs: Any) -> "DataProtocol":
        return cls.get(name)(**kwargs)

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global data_dict

        def before(cls_: Type) -> None:
            cls_.__identifier__ = name

        return register_core(name, data_dict, before_register=before)


class SamplerProtocol(ABC):
    shuffle: bool

    @abstractmethod
    def __init__(self, data: DataProtocol, **kwargs: Any):
        self.data = data

    @classmethod
    def get(cls, name: str) -> Type["SamplerProtocol"]:
        return sampler_dict[name]

    @classmethod
    def make(cls, name: str, data: DataProtocol, **kwargs: Any) -> "SamplerProtocol":
        return cls.get(name)(data, **kwargs)

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global sampler_dict

        def before(cls_: Type) -> None:
            cls_.__identifier__ = name

        return register_core(name, sampler_dict, before_register=before)


class DataLoaderProtocol(ABC):
    is_onnx: bool = False
    _num_siamese: int = 1
    labels_key: str = "labels"

    data: DataProtocol
    sampler: SamplerProtocol
    enabled_sampling: bool
    return_indices: bool
    batch_size: int
    _verbose_level: int

    @abstractmethod
    def __init__(
        self,
        batch_size: int,
        sampler: SamplerProtocol,
        *,
        return_indices: bool = False,
        verbose_level: int = 2,
        **kwargs: Any,
    ):
        self.batch_size = batch_size
        self.sampler = sampler
        self.return_indices = return_indices
        self._verbose_level = verbose_level

    @abstractmethod
    def __len__(self) -> int:
        pass

    @abstractmethod
    def __iter__(self) -> "DataLoaderProtocol":
        pass

    @abstractmethod
    def __next__(self) -> loader_batch_type:
        pass

    @abstractmethod
    def copy(self) -> "DataLoaderProtocol":
        pass

    @property
    def num_samples(self) -> int:
        return len(self.data)

    @classmethod
    def get(cls, name: str) -> Type["DataLoaderProtocol"]:
        return loader_dict[name]

    @classmethod
    def make(cls, name: str, *args: Any, **kwargs: Any) -> "DataLoaderProtocol":
        return cls.get(name)(*args, **kwargs)

    @classmethod
    def register(cls, name: str) -> Callable[[Type], Type]:
        global loader_dict

        def before(cls_: Type) -> None:
            cls_.__identifier__ = name

        return register_core(name, loader_dict, before_register=before)


class PrefetchLoader:
    def __init__(
        self,
        loader: DataLoaderProtocol,
        device: Union[str, torch.device],
        *,
        is_onnx: bool = False,
        enable_prefetch: bool = False,
    ):
        self.loader = loader
        self.device = device
        self.is_onnx = is_onnx
        loader.is_onnx = is_onnx
        self.enable_prefetch = enable_prefetch
        self.data = loader.data
        self.return_indices = loader.return_indices
        self.stream = None if not self.use_stream else torch.cuda.Stream(device)
        self.next_batch: Union[np_dict_type, tensor_dict_type]
        self.next_batch_indices: Optional[torch.Tensor]
        self.stop_at_next_batch = False
        self.batch_size = loader.batch_size
        self._num_siamese = loader._num_siamese

    def __len__(self) -> int:
        return len(self.loader)

    def __iter__(self) -> "PrefetchLoader":
        self.stop_at_next_batch = False
        self.loader.__iter__()
        self.preload()
        return self

    def __next__(self) -> prefetch_batch_type:
        if self.stop_at_next_batch:
            raise StopIteration
        if self.use_stream:
            torch.cuda.current_stream(self.device).wait_stream(self.stream)
        batch, batch_indices = self.next_batch, self.next_batch_indices
        self.preload()
        return batch, batch_indices

    def _to_device(self, indices_tensor: Optional[torch.Tensor]) -> None:
        if not self.enable_prefetch:
            kwargs = {}
        else:
            kwargs = {"non_blocking": self.enable_prefetch}
        self.next_batch = {
            k: None if v is None else v.to(self.device, **kwargs)  # type: ignore
            for k, v in self.next_batch.items()
        }
        if indices_tensor is None:
            self.next_batch_indices = None
        else:
            indices_tensor = indices_tensor.to(self.device, **kwargs)  # type: ignore
            self.next_batch_indices = indices_tensor

    def preload(self) -> None:
        try:
            sample = self.loader.__next__()
        except StopIteration:
            self.stop_at_next_batch = True
            return None
        indices_tensor: Optional[torch.Tensor]
        if not self.return_indices:
            indices_tensor = None
        else:
            sample, batch_indices = sample  # type: ignore
            indices_tensor = to_torch(batch_indices).to(torch.long)

        self.next_batch = sample  # type: ignore
        if self.is_cpu:
            self.next_batch_indices = indices_tensor
            return None

        if not self.enable_prefetch:
            self._to_device(indices_tensor)
        else:
            with torch.cuda.stream(self.stream):
                self._to_device(indices_tensor)

    @property
    def is_cpu(self) -> bool:
        if self.is_onnx:
            return True
        if isinstance(self.device, str):
            return self.device == "cpu"
        return self.device.type == "cpu"

    @property
    def use_stream(self) -> bool:
        return self.enable_prefetch and not self.is_cpu


class TrainerState:
    def __init__(self, trainer_config: Dict[str, Any]):
        # properties
        self.step = self.epoch = 0
        self.batch_size: int
        self.num_step_per_epoch: int
        self.log_disabled: bool = False
        # settings
        self.config = trainer_config
        self.min_epoch = self.config["min_epoch"]
        self.num_epoch = self.config["num_epoch"]
        self.max_epoch = self.config["max_epoch"]
        self.log_patience = self.config.setdefault("log_patience", None)
        self.max_snapshot_file = int(self.config.setdefault("max_snapshot_file", 5))
        self.min_num_sample = self.config.setdefault("min_num_sample", 3000)
        self._snapshot_start_step = self.config.setdefault("snapshot_start_step", None)
        num_step_per_snapshot = self.config.setdefault("num_step_per_snapshot", 0)
        num_snapshot_per_epoch = self.config.setdefault("num_snapshot_per_epoch", 2)
        max_step_per_snapshot = self.config.setdefault("max_step_per_snapshot", 1000)
        plateau_start = self.config.setdefault("plateau_start_snapshot", self.num_epoch)
        self._num_step_per_snapshot = int(num_step_per_snapshot)
        self.num_snapshot_per_epoch = int(num_snapshot_per_epoch)
        self.max_step_per_snapshot = int(max_step_per_snapshot)
        self.plateau_start = int(plateau_start)

    def inject_loader(self, loader: DataLoaderProtocol) -> None:
        self.batch_size = loader.batch_size
        self.num_step_per_epoch = len(loader)

    def set_terminate(self) -> None:
        self.step = self.epoch = -1

    @property
    def is_terminate(self) -> bool:
        return self.epoch == -1

    @property
    def snapshot_start_step(self) -> int:
        if self._snapshot_start_step is not None:
            return self._snapshot_start_step
        return int(math.ceil(self.min_num_sample / self.batch_size))

    @property
    def num_step_per_snapshot(self) -> int:
        if self._num_step_per_snapshot > 0:
            return self._num_step_per_snapshot
        return max(
            1,
            min(
                self.max_step_per_snapshot,
                int(self.num_step_per_epoch / self.num_snapshot_per_epoch),
            ),
        )

    @property
    def should_train(self) -> bool:
        return self.epoch < self.num_epoch

    @property
    def should_monitor(self) -> bool:
        return self.step % self.num_step_per_snapshot == 0

    @property
    def should_log_lr(self) -> bool:
        if self.log_disabled:
            return False
        denominator = min(self.num_step_per_epoch, 10)
        return self.step % denominator == 0

    @property
    def should_log_losses(self) -> bool:
        if self.log_disabled:
            return False
        if self.log_patience is not None:
            patience = self.log_patience
        else:
            patience = max(4, int(round(self.num_step_per_epoch / 50.0)))
        denominator = min(self.num_step_per_epoch, patience)
        return self.step % denominator == 0

    @property
    def should_log_artifacts(self) -> bool:
        return self.should_log_metrics_msg

    @property
    def should_log_metrics_msg(self) -> bool:
        if self.log_disabled:
            return False
        min_period = self.max_step_per_snapshot / 3
        min_period = math.ceil(min_period / self.num_step_per_snapshot)
        period = max(1, int(min_period)) * self.num_step_per_snapshot
        return self.step % period == 0

    @property
    def should_start_snapshot(self) -> bool:
        return self.step >= self.snapshot_start_step

    @property
    def should_start_monitor_plateau(self) -> bool:
        return self.step >= self.plateau_start * self.num_step_per_snapshot

    @property
    def should_extend_epoch(self) -> bool:
        return self.epoch == self.num_epoch and self.epoch < self.max_epoch

    @property
    def reached_max_epoch(self) -> bool:
        return self.epoch == self.max_epoch

    @property
    def disable_logging(self) -> context_error_handler:
        class _(context_error_handler):
            def __init__(self, state: TrainerState):
                self.state = state
                self.disabled = state.log_disabled

            def __enter__(self) -> None:
                self.state.log_disabled = True

            def _normal_exit(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                self.state.log_disabled = self.disabled

        return _(self)


class StepOutputs(NamedTuple):
    forward_results: tensor_dict_type
    loss_dict: tensor_dict_type
    loss_items: Dict[str, float]


class InferenceOutputs(NamedTuple):
    results: np_dict_type
    loss_items: Optional[Dict[str, float]]
    labels: Optional[np.ndarray]
    probabilities: Optional[np.ndarray]


# Protocols below are meant to fit with and only with `Trainer`


class TrainerDataProtocol(ABC):
    task_type: TaskTypes

    @abstractmethod
    def __len__(self) -> int:
        pass

    @property
    def is_reg(self) -> bool:
        return self.task_type.is_reg


class ModelProtocol(nn.Module, LoggingMixinWithRank, metaclass=ABCMeta):
    timing: bool = True
    pt_prefix: str = "model_"
    scores_file: str = "scores.json"

    __identifier__: str
    data: DataProtocol
    device: torch.device
    ema: Optional[EMA] = None
    num_train: Optional[int] = None
    num_valid: Optional[int] = None

    @property
    def use_ema(self) -> bool:
        return self.ema is not None

    def init_ema(self) -> None:
        if self.config is None:
            return None
        ema_decay = self.config.setdefault("ema_decay", 0.0)
        if 0.0 < ema_decay < 1.0:
            named_params = list(self.named_parameters())
            self.ema = EMA(ema_decay, named_params)  # type: ignore

    def apply_ema(self) -> None:
        if self.ema is None:
            raise ValueError("`ema` is not defined")
        self.ema()

    def info(self, *, return_only: bool = False) -> str:
        msg = "\n".join(["=" * 100, "configurations", "-" * 100, ""])
        msg += (
            pprint.pformat(self.configurations, compact=True) + "\n" + "-" * 100 + "\n"
        )
        msg += "\n".join(["=" * 100, "parameters", "-" * 100, ""])
        for name, param in self.named_parameters():
            if param.requires_grad:
                msg += name + "\n"
        msg += "\n".join(["-" * 100, "=" * 100, "buffers", "-" * 100, ""])
        for name, param in self.named_buffers():
            msg += name + "\n"
        msg += "\n".join(
            ["-" * 100, "=" * 100, "structure", "-" * 100, str(self), "-" * 100, ""]
        )
        if not return_only:
            self.log_block_msg(msg, verbose_level=4)  # type: ignore
        all_msg, msg = msg, "=" * 100 + "\n"
        msg += f"{self.info_prefix}training data : {self.num_train or 'unknown'}\n"
        msg += f"{self.info_prefix}valid    data : {self.num_valid or 'unknown'}\n"
        msg += "-" * 100
        if not return_only:
            self.log_block_msg(msg, verbose_level=3)  # type: ignore
        return "\n".join([all_msg, msg])

    def sorted_checkpoints(self, folder: str) -> List[str]:
        # better checkpoints will be placed earlier,
        #  which means `checkpoints[0]` is the best checkpoint
        scores_path = os.path.join(folder, self.scores_file)
        if not os.path.isfile(scores_path):
            return []
        with open(scores_path, "r") as f:
            scores = json.load(f)
        files = list(scores.keys())
        scores_list = [scores[file] for file in files]
        sorted_indices = np.argsort(scores_list)[::-1]
        return [files[i] for i in sorted_indices]

    def restore_checkpoint(
        self,
        folder: str,
        strict: bool = True,
        state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    ) -> bool:
        checkpoints = self.sorted_checkpoints(folder)
        if not checkpoints:
            self.log_msg(  # type: ignore
                f"no model file found in {folder}",
                self.warning_prefix,  # type: ignore
                msg_level=logging.WARNING,
            )
            return False
        success = False
        for checkpoint in checkpoints:
            model_file = os.path.join(folder, checkpoint)
            if not os.path.isfile(model_file):
                continue
            self.log_msg(  # type: ignore
                f"restoring from {model_file}",
                self.info_prefix,  # type: ignore
                4,
            )
            states = torch.load(model_file, map_location=self.device)
            if state_dict_callback is not None:
                state_dict_callback(states)
            self.load_state_dict(states, strict)
            success = True
            break
        return success

    def step(
        self,
        state: TrainerState,
        batch_idx: int,
        batch: tensor_dict_type,
        batch_indices: Optional[torch.Tensor],
        loader_name: str,
        engine: Optional[Any] = None,
    ) -> StepOutputs:
        with timing_context(self, "model.forward", enable=self.timing):
            args = batch, batch_idx, state, batch_indices, loader_name
            forward_results = (engine or self)(*args)
        with timing_context(self, "loss.forward", enable=self.timing):
            loss_dict = self.loss_function(
                batch_idx,
                batch,
                batch_indices,
                forward_results,
                state,
            )
        with timing_context(self, "loss.to_item", enable=self.timing):
            loss_items = {k: v.item() for k, v in loss_dict.items()}
        return StepOutputs(forward_results, loss_dict, loss_items)

    @property
    @abstractmethod
    def config(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def configurations(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def output_probabilities(self) -> bool:
        pass

    @abstractmethod
    def forward(
        self,
        batch: tensor_dict_type,
        batch_idx: Optional[int] = None,
        state: Optional[TrainerState] = None,
        batch_indices: Optional[np.ndarray] = None,
        loader_name: Optional[str] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        pass

    @abstractmethod
    def loss_function(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        batch_indices: Optional[torch.Tensor],
        forward_results: tensor_dict_type,
        state: Optional[TrainerState],
    ) -> tensor_dict_type:
        pass


class InferenceProtocol(ABC):
    data: DataProtocol
    model: Optional[ModelProtocol]
    is_binary: bool
    binary_metric: Optional[str]
    binary_threshold: Optional[float]
    use_binary_threshold: bool
    onnx: Any = None
    use_tqdm: bool = True
    use_grad_in_predict: bool = False

    @property
    def binary_config(self) -> Dict[str, Any]:
        return {
            "binary_metric": self.binary_metric,
            "binary_threshold": self.binary_threshold,
        }

    @property
    def output_probabilities(self) -> bool:
        if self.onnx is not None:
            return self.onnx.output_probabilities
        if self.model is None:
            raise ValueError("either `onnx` or `model` should be provided")
        return self.model.output_probabilities

    @property
    def need_binary_threshold(self) -> bool:
        if not self.use_binary_threshold:
            return False
        return self.is_binary and self.binary_metric is not None

    def to_tqdm(self, loader: PrefetchLoader) -> Union[tqdm, PrefetchLoader]:
        if not self.use_tqdm:
            return loader
        return tqdm(loader, total=len(loader), leave=False, position=2)

    # API

    def get_outputs(
        self,
        loader: PrefetchLoader,
        loader_name: Optional[str],
        *,
        use_tqdm: bool = False,
        return_loss: bool = True,
        return_outputs: bool = True,
        state: Optional[TrainerState] = None,
        portion: float = 1.0,
        **kwargs: Any,
    ) -> InferenceOutputs:
        labels_key = loader.loader.labels_key
        if use_tqdm:
            loader = self.to_tqdm(loader)

        def _core() -> InferenceOutputs:
            results: Dict[str, Optional[List[np.ndarray]]] = {}
            loss_items: Dict[str, List[float]] = {}
            labels = []
            for i, (batch, batch_indices) in enumerate(loader):
                if i / len(loader) >= portion:
                    break
                local_labels = batch[labels_key]
                if local_labels is not None:
                    if not isinstance(local_labels, np.ndarray):
                        local_labels = to_numpy(local_labels)
                    labels.append(local_labels)
                if self.onnx is not None:
                    local_results = self.onnx.inference(batch)
                    local_losses = None
                else:
                    assert self.model is not None
                    with eval_context(self.model, use_grad=use_grad):
                        assert not self.model.training
                        local_kwargs = shallow_copy_dict(kwargs)
                        local_kwargs["return_loss"] = return_loss
                        local_results = self.model(
                            batch,
                            i,
                            state,
                            batch_indices,
                            loader_name,
                            **local_kwargs,
                        )
                    if not return_loss:
                        local_losses = None
                    else:
                        with eval_context(self.model, use_grad=use_grad):
                            assert not self.model.training
                            local_losses = self.model.loss_function(
                                i,
                                batch,
                                batch_indices,
                                local_results,
                                state,
                            )
                for k, v in local_results.items():
                    if v is None:
                        continue
                    if self.onnx is not None:
                        v_np = v
                    else:
                        v_np = to_numpy(v)
                    if not return_outputs:
                        results[k] = None
                    else:
                        results.setdefault(k, []).append(v_np)  # type: ignore
                if local_losses is not None:
                    for k, v in local_losses.items():
                        loss_items.setdefault(k, []).append(v.item())

            if return_outputs:
                results = {k: np.vstack(v) for k, v in results.items()}

            return InferenceOutputs(
                results,
                None
                if not loss_items
                else {k: sum(v) / len(v) for k, v in loss_items.items()},
                None if not labels else np.vstack(labels),
                None,
            )

        use_grad = kwargs.pop("use_grad", self.use_grad_in_predict)
        try:
            return _core()
        except:
            use_grad = self.use_grad_in_predict = True
            return _core()

    def predict_from_outputs(
        self,
        outputs: InferenceOutputs,
        return_all: bool = False,
        requires_recover: bool = True,
        returns_probabilities: bool = False,
        **kwargs: Any,
    ) -> Union[np.ndarray, np_dict_type]:
        results = outputs.results
        # regression
        if self.data.is_reg:
            return_key = kwargs.get("return_key", "predictions")
            fn = partial(self.data.recover_labels, inplace=True)
            if not return_all:
                predictions = results[return_key]
                if requires_recover:
                    if predictions.shape[1] == 1:
                        return fn(predictions)
                    return np.apply_along_axis(fn, axis=0, arr=predictions).squeeze()
                return predictions
            if not requires_recover:
                return results
            recovered = {}
            for k, v in results.items():
                if is_float(v):
                    if v.shape[1] == 1:
                        v = fn(v)
                    else:
                        v = np.apply_along_axis(fn, axis=0, arr=v).squeeze()
                recovered[k] = v
            return recovered

        # classification
        def _return(new_predictions: np.ndarray) -> Union[np.ndarray, np_dict_type]:
            if not return_all:
                return new_predictions
            results["predictions"] = new_predictions
            return results

        predictions = results["logits"] = results["predictions"]
        if returns_probabilities:
            if not self.output_probabilities:
                predictions = to_prob(predictions)
            return _return(predictions)
        if not self.is_binary or self.binary_threshold is None:
            return _return(predictions.argmax(1).reshape([-1, 1]))

        if self.output_probabilities:
            probabilities = predictions
        else:
            probabilities = to_prob(predictions)
        return _return(self.predict_with(probabilities))

    def predict(
        self,
        loader: PrefetchLoader,
        *,
        return_all: bool = False,
        requires_recover: bool = True,
        returns_probabilities: bool = False,
        loader_name: Optional[str] = None,
        return_loss: bool = False,
        use_tqdm: bool = False,
        **kwargs: Any,
    ) -> Union[np.ndarray, np_dict_type]:
        # Notice : when `return_all` is True,
        #  there might not be `predictions` key in the results
        outputs = self.get_outputs(
            loader,
            loader_name,
            use_tqdm=use_tqdm,
            return_loss=return_loss,
            **shallow_copy_dict(kwargs),
        )
        return self.predict_from_outputs(
            outputs,
            return_all,
            requires_recover,
            returns_probabilities,
            **shallow_copy_dict(kwargs),
        )

    def predict_with(self, probabilities: np.ndarray) -> np.ndarray:
        if not self.is_binary or self.binary_threshold is None:
            return probabilities.argmax(1).reshape([-1, 1])
        predictions = (
            (probabilities[..., 1] >= self.binary_threshold)
            .astype(np_int_type)
            .reshape([-1, 1])
        )
        return predictions

    def generate_binary_threshold(
        self,
        loader: PrefetchLoader,
        loader_name: str,
        *,
        return_loss: bool = True,
        use_tqdm: bool = False,
    ) -> Optional[InferenceOutputs]:
        if not self.need_binary_threshold:
            return None
        outputs = self.get_outputs(
            loader,
            loader_name,
            return_loss=return_loss,
            getting_metrics=True,
            use_tqdm=use_tqdm,
        )
        results = self.predict_from_outputs(
            outputs,
            return_all=True,
            returns_probabilities=True,
        )
        labels = outputs.labels
        probabilities = results["predictions"]
        try:
            threshold = Metrics.get_binary_threshold(
                labels,
                probabilities,
                self.binary_metric,
            )
            self.binary_threshold = threshold.item()
        except ValueError:
            self.binary_threshold = None

        new_outputs = InferenceOutputs(*outputs[:3], probabilities)
        return None if loader_name == "tr" else new_outputs


__all__ = [
    "DataSplit",
    "DataProtocol",
    "SamplerProtocol",
    "DataLoaderProtocol",
    "PrefetchLoader",
    "TrainerState",
    "StepOutputs",
    "InferenceOutputs",
    "TrainerDataProtocol",
    "ModelProtocol",
    "InferenceProtocol",
]



import os
import json
import math

import numpy as np

from typing import Any
from typing import Dict
from typing import NamedTuple
from argparse import Namespace
from cftool.misc import update_dict
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.misc import LoggingMixin
from cflearn.types import data_type
from cflearn.types import general_config_type

from ..configs import _parse_config
from ..pipeline import Pipeline

cli_root = os.path.join(os.path.dirname(__file__), os.pardir, os.pardir, "cli")
cli_root = os.path.abspath(cli_root)


class DeepspeedResults(NamedTuple):
    m: Pipeline
    workplace: str
    config_path: str
    ds_config_path: str


def impute_deepspeed_args(
    args: Namespace,
    config: Dict[str, Any],
    trainer_config: Dict[str, Any],
) -> None:
    if args.deepspeed_config is not None:
        config["use_tqdm"] = False
        config["trigger_logging"] = True
        config["ds_args"] = args
        with open(args.deepspeed_config, "r") as f:
            ds_config = json.load(f)
        ds_config_changed = False
        if trainer_config["use_amp"]:
            ds_config_changed = True
            ds_config["fp16"] = {
                "enabled": True,
                "loss_scale": 0,
                "initial_scale_power": 32,
                "loss_scale_window": 1000,
                "hysteresis": 2,
                "min_loss_scale": 1,
            }
        clip_norm = trainer_config["clip_norm"]
        if clip_norm > 0.0:
            ds_config_changed = True
            ds_config["gradient_clipping"] = clip_norm
        if ds_config_changed:
            folder, file = os.path.split(args.deepspeed_config)
            name, ext = os.path.splitext(file)
            new_file = f"_cf_{name}{ext}"
            new_path = new_file if not folder else os.path.join(folder, new_file)
            with lock_manager(folder or "./", new_file):
                with open(new_path, "w") as f:
                    json.dump(ds_config, f)
            args.deepspeed_config = new_path


def impute_deepspeed_config(
    cuda: str,
    default_batch_size: int,
    ds_config: general_config_type,
    final_config: Dict[str, Any],
) -> Dict[str, Any]:
    parsed_ds_config = _parse_config(ds_config)
    num_cuda = len(cuda.split(","))
    batch_size = parsed_ds_config.setdefault("train_batch_size", default_batch_size)
    micro_batch_size = 2 ** (math.floor(math.log2(batch_size / num_cuda)))
    parsed_ds_config["train_micro_batch_size_per_gpu"] = micro_batch_size
    new_batch_size = micro_batch_size * num_cuda
    if batch_size != new_batch_size:
        print(
            f"{LoggingMixin.warning_prefix}`batch_size` "
            f"will be switched from {batch_size} to {new_batch_size}"
        )
        parsed_ds_config["train_batch_size"] = new_batch_size
        batch_size = new_batch_size
    parsed_ds_config.setdefault("steps_per_print", 30000 // batch_size)
    final_config["batch_size"] = batch_size
    return parsed_ds_config


def deepspeed(
    x: data_type,
    y: data_type = None,
    x_cv: data_type = None,
    y_cv: data_type = None,
    *,
    model: str = "fcnn",
    workplace: str = "__deepspeed__",
    config: general_config_type = None,
    ds_config: general_config_type = None,
    increment_config: general_config_type = None,
    cuda: str = "0",
    **kwargs: Any,
) -> DeepspeedResults:
    os.makedirs(workplace, exist_ok=True)
    # logging
    default_logging_folder = os.path.join(workplace, "_logs")
    kwargs.setdefault("logging_folder", default_logging_folder)
    # data
    data_folder = os.path.join(workplace, "__data__")
    os.makedirs(data_folder, exist_ok=True)
    if not isinstance(x, np.ndarray):
        train_file = x
        valid_file = x_cv
    else:
        msg_base = "`{}` should be provided when `{}` is a numpy array"
        if y is None:
            raise ValueError(msg_base.format("y", "x"))
        train_file = os.path.join(data_folder, "train.npy")
        np.save(train_file, np.hstack([x, y]))
        valid_file = None
        if x_cv is not None:
            if y_cv is None:
                raise ValueError(msg_base.format("y_cv", "x_cv"))
            valid_file = os.path.join(data_folder, "valid.npy")
            np.save(valid_file, np.hstack([x_cv, y_cv]))
    # config
    parsed_config = update_dict(_parse_config(config), kwargs)
    parsed_increment_config = _parse_config(increment_config)
    final_config = update_dict(parsed_increment_config, parsed_config)
    final_config["model_saving_folder"] = workplace
    # deepspeed
    parsed_ds_config = impute_deepspeed_config(cuda, 256, ds_config, final_config)
    # save configs
    config_file = Saving.save_dict(final_config, "config", workplace)
    ds_config_file = Saving.save_dict(parsed_ds_config, "ds_config", workplace)

    os.system(
        f"deepspeed "
        f"--include='localhost:{cuda}' "
        f"{os.path.join(cli_root, 'main.py')} "
        f"--model {model} "
        f"--config {config_file} "
        f"--train_file {train_file} "
        f"{'' if valid_file is None else f'--valid_file {valid_file} '}"
        f"--deepspeed_config {ds_config_file}"
    )
    return DeepspeedResults(
        Pipeline.load(os.path.join(workplace, "pipeline")),
        workplace,
        config_file,
        ds_config_file,
    )


__all__ = [
    "impute_deepspeed_args",
    "deepspeed",
]



import os
import shutil

from typing import *
from torch import Tensor
from cftool.misc import shallow_copy_dict
from cftool.misc import LoggingMixin
from cftool.ml.utils import register_metric
from cfdata.tabular.processors.base import Processor

from ..modules import *
from ..types import tensor_dict_type
from ..losses import LossBase
from ..configs import Configs
from ..misc.toolkit import Initializer
from ..models.base import model_dict
from ..models.base import ModelBase
from ..modules.heads import head_dict
from ..modules.heads import HeadBase
from ..modules.heads import HeadConfigs
from ..modules.extractors import ExtractorBase
from ..modules.aggregators import AggregatorBase


def register_initializer(name: str) -> Callable[[Callable], Callable]:
    def _register(f: Callable) -> Callable:
        Initializer.add_initializer(f, name)
        return f

    return _register


def register_processor(name: str) -> Callable[[Type], Type]:
    return Processor.register(name)


def register_loss(name: str) -> Callable[[Type], Type]:
    return LossBase.register(name)


def register_extractor(name: str) -> Callable[[Type], Type]:
    return ExtractorBase.register(name)


def register_head(name: str) -> Callable[[Type], Type]:
    return HeadBase.register(name)


def register_aggregator(name: str) -> Callable[[Type], Type]:
    return AggregatorBase.register(name)


class PipeInfo(NamedTuple):
    key: str
    transform: str = "default"
    extractor: Optional[str] = None
    head: Optional[str] = None
    extractor_config: str = "default"
    head_config: str = "default"
    extractor_meta_scope: Optional[str] = None
    head_meta_scope: Optional[str] = None


def register_pipe(
    key: str,
    *,
    transform: str = "default",
    extractor: Optional[str] = None,
    head: Optional[str] = None,
    extractor_config: str = "default",
    head_config: str = "default",
    extractor_meta_scope: Optional[str] = None,
    head_meta_scope: Optional[str] = None,
) -> Callable[[Type], Type]:
    return ModelBase.register_pipe(
        key,
        transform=transform,
        extractor=extractor,
        head=head,
        extractor_config=extractor_config,
        head_config=head_config,
        extractor_meta_scope=extractor_meta_scope,
        head_meta_scope=head_meta_scope,
    )


def register_model(
    name: str,
    *,
    pipes: Optional[List[PipeInfo]] = None,
) -> Optional[Callable[[Type], Type]]:
    if pipes is None:
        return ModelBase.register(name)

    @ModelBase.register(name)
    class _(ModelBase):
        pass

    for pipe in pipes:
        _ = register_pipe(  # type: ignore
            pipe.key,
            transform=pipe.transform,
            extractor=pipe.extractor,
            head=pipe.head,
            extractor_config=pipe.extractor_config,
            head_config=pipe.head_config,
            extractor_meta_scope=pipe.extractor_meta_scope,
            head_meta_scope=pipe.head_meta_scope,
        )(_)

    return None


def _register_config(
    base: Type[Configs],
    scope: str,
    name: str,
    config: Optional[Dict[str, Any]] = None,
) -> Optional[Callable[[Type], Type]]:
    if config is None:
        return base.register(scope, name)

    def register(config_: Dict[str, Any]) -> None:
        @base.register(scope, name)
        class _(base):  # type: ignore
            def get_default(self) -> Dict[str, Any]:
                return shallow_copy_dict(config_)

    register(config)
    return None


def register_config(
    scope: str,
    name: str,
    *,
    config: Optional[Dict[str, Any]] = None,
) -> Optional[Callable[[Type], Type]]:
    return _register_config(Configs, scope, name, config)  # type: ignore


def register_head_config(
    scope: str,
    name: str,
    *,
    head_config: Optional[Dict[str, Any]] = None,
) -> Optional[Callable[[Type], Type]]:
    return _register_config(HeadConfigs, scope, name, head_config)  # type: ignore


api_root = os.path.abspath(os.path.dirname(__file__))
external_root = os.path.join(api_root, os.pardir, "external_")
external_root = os.path.abspath(external_root)
external_init_file = os.path.join(external_root, "__init__.py")


def _get_external_files() -> List[str]:
    files = []
    for file in sorted(os.listdir(external_root)):
        if not os.path.isfile(os.path.join(external_root, file)):
            continue
        if file == "__init__.py":
            continue
        if os.path.splitext(file)[1] != ".py":
            continue
        files.append(file)
    return files


def _refresh_external_init() -> None:
    with open(external_init_file, "w") as f:
        for file in _get_external_files():
            f.write(f"from .{os.path.splitext(file)[0]} import *\n")


def register_external(file_path: str) -> None:
    file = os.path.basename(file_path)
    tgt_path = os.path.join(external_root, file)
    if os.path.isfile(tgt_path):
        print(
            f"{LoggingMixin.warning_prefix}'{file}' already registered, "
            "it will be overwritten."
        )
    shutil.copy(file_path, tgt_path)
    _refresh_external_init()
    import cflearn
    import importlib

    importlib.reload(cflearn)
    importlib.reload(cflearn.external_)
    exec(f"importlib.reload(cflearn.external_.{os.path.splitext(file)[0]})")


def remove_external(file_path: str) -> None:
    file = os.path.basename(file_path)
    tgt_path = os.path.join(external_root, file)
    if not os.path.isfile(tgt_path):
        raise ValueError(f"'{file}' is not registered")
    os.remove(tgt_path)
    _refresh_external_init()


def remove_all_external() -> None:
    for file in _get_external_files():
        remove_external(os.path.join(external_root, file))


def register_module(name: str, module_base: Any) -> None:
    current_head = head_dict.pop(name, None)
    current_model = model_dict.pop(name, None)
    if current_head is not None:
        print(
            f"{LoggingMixin.warning_prefix}'{name}' already exists in `head_dict`, "
            "it will be overwritten"
        )
    if current_model is not None:
        print(
            f"{LoggingMixin.warning_prefix}'{name}' already exists in `model_dict`, "
            "it will be overwritten"
        )

    @register_head(name)
    class _(HeadBase):
        def __init__(self, in_dim: int, out_dim: int, **kwargs: Any):
            super().__init__(in_dim, out_dim, **kwargs)
            self.model = module_base(in_dim, out_dim, **kwargs)

        def forward(self, net: Tensor) -> Union[Tensor, tensor_dict_type]:
            return self.model(net)

    register_head_config(name, "default", head_config={})
    register_model(name, pipes=[PipeInfo(name, extractor="identity_ts")])


__all__ = [
    "register_extractor",
    "register_head",
    "register_aggregator",
    "register_pipe",
    "register_model",
    "register_config",
    "register_head_config",
    "register_metric",
    "register_optimizer",
    "register_scheduler",
    "register_initializer",
    "register_processor",
    "register_loss",
    "register_external",
    "remove_external",
    "remove_all_external",
    "register_module",
    "Initializer",
    "Processor",
    "LossBase",
    "PipeInfo",
]



from .base import ModelBase


@ModelBase.register("tree_dnn")
@ModelBase.register_pipe("dndf")
@ModelBase.register_pipe("fcnn", transform="embedding", head_config="pruned")
class TreeDNN(ModelBase):
    pass


@ModelBase.register("tree_stack")
@ModelBase.register_pipe("tree_stack")
class TreeStack(ModelBase):
    @property
    def output_probabilities(self) -> bool:
        return True


@TreeStack.register("tree_linear")
@ModelBase.register_pipe("tree_stack", head_config="linear")
class TreeLinear(TreeStack):
    pass


__all__ = ["TreeDNN", "TreeStack", "TreeLinear"]



import torch

from typing import *
from torch.nn.functional import l1_loss
from torch.nn.functional import softplus

from ...losses import LossBase
from ...types import losses_type
from ...types import tensor_dict_type
from ...misc.toolkit import LoggingMixinWithRank
from ...modules.auxiliary import MTL


@LossBase.register("ddr")
class DDRLoss(LossBase, LoggingMixinWithRank):
    def _init_config(self, config: Dict[str, Any]) -> None:
        self.fetch_q = config["fetch_q"]
        self.fetch_cdf = config["fetch_cdf"]
        self.mtl = MTL(18, config["mtl_method"])
        self._lb_pdf = config.setdefault("lambda_pdf", 0.01)
        self._pdf_eps = config.setdefault("pdf_eps", 1.0e-12)
        self._lb_recover = config.setdefault("lambda_recover", 1.0)

    @staticmethod
    def _median_losses(
        forward_results: tensor_dict_type,
        target: torch.Tensor,
    ) -> Tuple[torch.Tensor, tensor_dict_type]:
        # median
        median = forward_results["predictions"]
        median_losses = l1_loss(median, target, reduction="none")
        # median residual
        pos_med_res = forward_results["med_pos_med_res"]
        neg_med_res = forward_results["med_neg_med_res"]
        target_median_residual = target - median.detach()
        tmr_sign_mask = torch.sign(target_median_residual) > 0
        mr_prediction = torch.where(tmr_sign_mask, pos_med_res, -neg_med_res)
        mr_losses = torch.abs(target_median_residual - mr_prediction)
        losses_dict = {"median": median_losses, "median_residual": mr_losses}
        return target_median_residual, losses_dict

    def _q_losses(
        self,
        forward_results: tensor_dict_type,
        target_median_residual: Optional[torch.Tensor],
        is_synthetic: bool,
    ) -> tensor_dict_type:
        # median residual anchor
        if is_synthetic:
            syn_med_mul = forward_results["syn_med_mul"].abs()
            return {"median_residual_anchor": (syn_med_mul.abs() - 1.0).abs()}
        # quantile
        q_batch = forward_results["q_batch"]
        assert q_batch is not None
        y_res = forward_results["y_res"]
        assert y_res is not None and target_median_residual is not None
        quantile_losses = self._quantile_losses(y_res, target_median_residual, q_batch)
        return {"quantile": quantile_losses}

    def _y_losses(
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        is_synthetic: bool,
    ) -> tensor_dict_type:
        # median residual anchor
        if is_synthetic:
            syn_cdf_logit_mul = forward_results["syn_cdf_logit_mul"]
            syn_med_cdf_logit_mul = forward_results["syn_med_cdf_logit_mul"]
            res_losses = (syn_cdf_logit_mul.abs() - 1.0).abs()
            med_losses = syn_med_cdf_logit_mul.abs()
            return {"cdf_anchor": res_losses + med_losses}
        # cdf
        y_batch = forward_results["y_batch"]
        cdf_logit = forward_results["cdf_logit"]
        assert y_batch is not None and cdf_logit is not None
        cdf_losses = self._cdf_losses(cdf_logit, target, y_batch)
        # pdf
        pdf = forward_results["pdf"]
        pdf_losses = None if pdf is None else self._pdf_losses(pdf)
        # combine
        losses = {"cdf": cdf_losses}
        if pdf_losses is not None:
            losses["pdf"] = pdf_losses
        return losses

    def _dual_losses(
        self,
        forward_results: tensor_dict_type,
        is_synthetic: bool,
    ) -> tensor_dict_type:
        # q recover
        q_batch = forward_results["q_batch"]
        q_inverse = forward_results["q_inverse"]
        q_recover_losses = l1_loss(q_inverse, q_batch, reduction="none")
        q_recover_losses = self._lb_recover * q_recover_losses
        # y recover
        y_batch = forward_results["y_batch"]
        y_recover_losses = None
        if not is_synthetic:
            median = forward_results["median"].detach()
            y_inverse = forward_results["y_inverse_res"] + median
            if y_inverse is not None:
                y_recover_losses = l1_loss(y_inverse, y_batch, reduction="none")
                y_recover_losses = self._lb_recover * y_recover_losses
        # combine
        losses = {"q_recover": q_recover_losses}
        if y_recover_losses is not None:
            losses["y_recover"] = y_recover_losses
        return losses

    def _core(  # type: ignore
        self,
        forward_results: tensor_dict_type,
        target: torch.Tensor,
        is_synthetic: bool = False,
    ) -> tensor_dict_type:
        losses: tensor_dict_type
        if is_synthetic:
            tmr, losses = None, {}
        else:
            tmr, losses = self._median_losses(forward_results, target)
        if self.fetch_q:
            losses.update(self._q_losses(forward_results, tmr, is_synthetic))
        if self.fetch_cdf:
            losses.update(self._y_losses(forward_results, target, is_synthetic))
        if self.fetch_q and self.fetch_cdf:
            losses.update(self._dual_losses(forward_results, is_synthetic))
        suffix = "" if not is_synthetic else "synthetic_"
        losses = {f"{suffix}{k}": v for k, v in losses.items()}
        # mtl
        if not self.mtl.registered:
            self.mtl.register(list(losses.keys()))
        losses["loss"] = self.mtl(losses)
        return losses

    def forward(  # type: ignore
        self,
        predictions: tensor_dict_type,
        target: torch.Tensor,
        *,
        is_synthetic: bool = False,
    ) -> tensor_dict_type:
        losses_dict = self._core(predictions, target, is_synthetic)
        reduced_losses_dict = {k: self._reduce(v) for k, v in losses_dict.items()}  # type: ignore
        return reduced_losses_dict

    @staticmethod
    def _quantile_losses(
        residual: torch.Tensor,
        target_residual: torch.Tensor,
        q_batch: torch.Tensor,
    ) -> torch.Tensor:
        quantile_error = target_residual - residual
        q1 = q_batch * quantile_error
        q2 = (q_batch - 1.0) * quantile_error
        return torch.max(q1, q2)

    @staticmethod
    def _cdf_losses(
        cdf_logit: torch.Tensor,
        target: torch.Tensor,
        y_batch: torch.Tensor,
    ) -> torch.Tensor:
        indicative = (target <= y_batch).to(torch.float32)
        return -indicative * cdf_logit + softplus(cdf_logit)

    def _pdf_losses(self, pdf: torch.Tensor) -> torch.Tensor:
        negative_mask = pdf <= self._pdf_eps
        losses = torch.zeros_like(pdf)
        losses[negative_mask] = -pdf[negative_mask]
        positive_mask = ~negative_mask
        losses[positive_mask] = -self._lb_pdf * torch.log(pdf[positive_mask])
        return losses


__all__ = ["DDRLoss"]



import math
import torch

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from abc import abstractmethod
from abc import ABCMeta
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from torch.nn import Module
from torch.nn import ModuleList
from cftool.misc import shallow_copy_dict
from cftool.misc import LoggingMixin
from cfdata.types import np_int_type
from cfdata.types import np_float_type

from .auxiliary import *
from ..misc.toolkit import *
from ..types import tensor_tuple_type


class Linear(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        bias: bool = True,
        pruner_config: Optional[Dict[str, Any]] = None,
        init_method: Optional[str] = "xavier_uniform",
        **kwargs: Any,
    ):
        super().__init__()
        self.linear = nn.Linear(in_dim, out_dim, bias)
        if pruner_config is None:
            pruner = None
        else:
            pruner = Pruner(pruner_config, [out_dim, in_dim])
        self.config, self.pruner = shallow_copy_dict(kwargs), pruner
        self._use_bias, self._init_method = bias, init_method
        with torch.no_grad():
            self.reset_parameters()

    @property
    def weight(self) -> Tensor:
        return self.linear.weight

    @property
    def bias(self) -> Optional[Tensor]:
        return self.linear.bias

    def forward(self, net: Tensor) -> Tensor:
        if self.pruner is None:
            return self.linear(net)
        weight = self.pruner(self.linear.weight)
        return F.linear(net, weight, self.linear.bias)

    def reset_parameters(self) -> None:
        if self._init_method is None:
            return
        if self._init_method not in Initializer.defined_initialization:
            return
        initializer = Initializer(self.config.setdefault("initialize_config", {}))
        assert isinstance(self.linear.weight, nn.Parameter)
        initializer.initialize(self.linear.weight, self._init_method)
        bias_fill = self.config.setdefault("bias_fill", 0.0)
        if self._use_bias:
            self.linear.bias.data.fill_(bias_fill)


class Mapping(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        bias: Optional[bool] = None,
        pruner_config: Optional[dict] = None,
        dropout: float = 0.5,
        batch_norm: bool = True,
        activation: Optional[str] = "ReLU",
        init_method: str = "xavier_uniform",
        **kwargs: Any,
    ):
        super().__init__()
        self.config = shallow_copy_dict(kwargs)
        if bias is None:
            bias = not batch_norm
        self.linear = Linear(
            in_dim,
            out_dim,
            bias=bias,
            pruner_config=pruner_config,
            init_method=init_method,
            **shallow_copy_dict(kwargs),
        )
        self.bn = None if not batch_norm else BN(out_dim)
        if activation is None:
            self.activation: Optional[Module] = None
        else:
            activation_config = self.config.setdefault("activation_config", None)
            self.activation = Activations.make(activation, activation_config)
        use_dropout = 0.0 < dropout < 1.0
        self.dropout = None if not use_dropout else Dropout(dropout)

    @property
    def weight(self) -> Tensor:
        return self.linear.weight

    @property
    def bias(self) -> Optional[Tensor]:
        return self.linear.bias

    def forward(self, net: Tensor, *, reuse: bool = False) -> Tensor:
        net = self.linear(net)
        if self.bn is not None:
            net = self.bn(net)
        if self.activation is not None:
            net = self.activation(net)
        if self.dropout is not None:
            net = self.dropout(net, reuse=reuse)
        return net

    @classmethod
    def simple(
        cls,
        in_dim: int,
        out_dim: int,
        *,
        bias: bool = False,
        dropout: float = 0.0,
        batch_norm: bool = False,
        activation: Optional[str] = None,
        pruner_config: Optional[Dict[str, Any]] = None,
    ) -> "Mapping":
        if activation != "glu":
            activation_config = {}
        else:
            activation_config = {"in_dim": out_dim, "bias": bias}
        return cls(
            in_dim,
            out_dim,
            bias=bias,
            pruner_config=pruner_config,
            dropout=dropout,
            batch_norm=batch_norm,
            activation=activation,
            activation_config=activation_config,
        )


class ResBlock(Module):
    def __init__(
        self,
        in_dim: int,
        latent_dim: int,
        *,
        bias: Optional[bool] = None,
        pruner_config: Optional[dict] = None,
        dropout: float = 0.0,
        batch_norm: bool = True,
        activation: Optional[str] = "ReLU",
        init_method: str = "xavier_uniform",
        **kwargs: Any,
    ):
        super().__init__()
        self.in_bn = None if not batch_norm else BN(in_dim)
        if activation is None:
            self.in_activation: Optional[Module] = None
        else:
            activation_config = kwargs.setdefault("activation_config", None)
            self.in_activation = Activations.make(activation, activation_config)
        self.mapping = Mapping(
            in_dim,
            latent_dim,
            bias=bias,
            pruner_config=pruner_config,
            dropout=dropout,
            batch_norm=batch_norm,
            activation=activation,
            init_method=init_method,
            **kwargs,
        )
        if bias is None:
            bias = True
        self.linear = Linear(
            latent_dim,
            latent_dim,
            bias=bias,
            pruner_config=pruner_config,
            init_method=init_method,
            **kwargs,
        )
        self.res_linear = Linear(
            in_dim,
            latent_dim,
            bias=bias,
            pruner_config=pruner_config,
            init_method=init_method,
            **kwargs,
        )

    def forward(self, net: Tensor) -> Tensor:
        latent = net
        if self.in_bn is not None:
            latent = self.in_bn(latent)
        if self.in_activation is not None:
            latent = self.in_activation(latent)
        latent = self.mapping(latent)
        latent = self.linear(latent)
        return self.res_linear(net) + latent


class MLP(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: Optional[int],
        num_units: List[int],
        mapping_configs: Union[Dict[str, Any], List[Dict[str, Any]]],
        *,
        final_mapping_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        mappings: List[Union[Linear, Mapping]] = []
        if isinstance(mapping_configs, dict):
            mapping_configs = [mapping_configs] * len(num_units)
        for num_unit, mapping_config in zip(num_units, mapping_configs):
            mappings.append(Mapping(in_dim, num_unit, **mapping_config))
            in_dim = num_unit
        if out_dim is not None:
            if final_mapping_config is None:
                final_mapping_config = {}
            mappings.append(Linear(in_dim, out_dim, **final_mapping_config))
        self.mappings = ModuleList(mappings)

    @property
    def weights(self) -> List[Tensor]:
        return [mapping.weight for mapping in self.mappings]

    @property
    def biases(self) -> List[Optional[Tensor]]:
        return [mapping.bias for mapping in self.mappings]

    def forward(self, net: Tensor) -> Tensor:
        for mapping in self.mappings:
            net = mapping(net)
        return net

    @classmethod
    def simple(
        cls,
        in_dim: int,
        out_dim: Optional[int],
        num_units: List[int],
        *,
        bias: bool = False,
        dropout: float = 0.0,
        batch_norm: bool = False,
        activation: Optional[str] = None,
        pruner_config: Optional[Dict[str, Any]] = None,
    ) -> "MLP":
        mapping_config: Dict[str, Any]
        mapping_config = {
            "bias": bias,
            "dropout": dropout,
            "batch_norm": batch_norm,
            "pruner_config": pruner_config,
        }
        if activation is not None:
            mapping_config["activation"] = activation
        mapping_configs: Union[Dict[str, Any], List[Dict[str, Any]]]
        if activation != "glu":
            mapping_configs = mapping_config
        else:
            mapping_configs = []
            for num_unit in num_units:
                cfg = shallow_copy_dict(mapping_config)
                cfg["activation_config"] = {"in_dim": num_unit, "bias": bias}
                mapping_configs.append(cfg)
        final_mapping_config = {"bias": bias}
        return cls(
            in_dim,
            out_dim,
            num_units,
            mapping_configs,
            final_mapping_config=final_mapping_config,
        )

    @staticmethod
    def get_funnel_settings(
        max_dropout: float,
        max_dim: int,
        out_dim: int,
        num_layers: int,
    ) -> Tuple[List[float], List[int]]:
        dim = None
        dim_decrease = int(math.ceil((max_dim - out_dim) / num_layers))
        drop_decrease = max_dropout / num_layers
        dropouts = []
        num_units = []
        current_drop = max_dropout
        for i in range(num_layers):
            dropouts.append(current_drop)
            current_drop -= drop_decrease
            if i == 0:
                dim = max_dim
            else:
                assert dim is not None
                dim -= dim_decrease
            num_units.append(dim)
        return dropouts, num_units

    @classmethod
    def funnel(
        cls,
        in_dim: int,
        out_dim: int,
        max_dim: int,
        num_layers: int,
        *,
        bias: Optional[bool] = None,
        max_dropout: float = 0.1,
        batch_norm: bool = False,
        activation: Optional[str] = "relu",
        pruner_config: Optional[Dict[str, Any]] = None,
        final_mapping_config: Optional[Dict[str, Any]] = None,
    ) -> "MLP":
        dropouts, num_units = cls.get_funnel_settings(
            max_dropout,
            max_dim,
            out_dim,
            num_layers,
        )
        mapping_configs = []
        for dropout in dropouts:
            mapping_config: Dict[str, Any] = {
                "bias": bias,
                "dropout": dropout,
                "batch_norm": batch_norm,
                "pruner_config": pruner_config,
            }
            if activation is not None:
                mapping_config["activation"] = activation
            mapping_configs.append(mapping_config)
        return cls(
            in_dim,
            out_dim,
            num_units,
            mapping_configs,
            final_mapping_config=final_mapping_config,
        )

    @classmethod
    def res_funnel(
        cls,
        in_dim: int,
        out_dim: int,
        max_dim: int,
        num_layers: int,
        *,
        bias: bool = True,
        max_dropout: float = 0.1,
        batch_norm: bool = True,
        activation: Optional[str] = "relu",
        pruner_config: Optional[Dict[str, Any]] = None,
    ) -> nn.Sequential:
        dropouts, num_units = cls.get_funnel_settings(
            max_dropout,
            max_dim,
            out_dim,
            num_layers,
        )
        in_linear = Linear(in_dim, max_dim, bias=bias, pruner_config=pruner_config)
        blocks: List[Module] = [in_linear]
        dim = max_dim
        for dropout, num_unit in zip(dropouts[1:], num_units[1:]):
            blocks.append(
                ResBlock(
                    dim,
                    num_unit,
                    bias=bias,
                    pruner_config=pruner_config,
                    dropout=dropout,
                    batch_norm=batch_norm,
                    activation=activation,
                )
            )
            dim = num_unit

        blocks.append(BN(dim))
        if activation is not None:
            blocks.append(Activations.make(activation))
        blocks.append(Linear(dim, out_dim, bias=bias, pruner_config=pruner_config))
        return nn.Sequential(*blocks)


class LeafAggregation(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Tensor:
        net, leaves = args
        softmax_leaves = F.softmax(leaves, dim=1)
        ctx.save_for_backward(net, softmax_leaves.t())
        return net.mm(softmax_leaves)

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Tuple[Optional[Tensor], ...]:
        grad_output = grad_outputs[0]
        if grad_output is None:
            return None, None
        net, softmax_leaves = ctx.saved_tensors
        net_grad = grad_output.mm(softmax_leaves)
        sub_grad = grad_output.t().mm(net)
        sub_grad2 = (softmax_leaves * sub_grad).sum(0, keepdim=True)
        leaves_grad = softmax_leaves * (sub_grad - sub_grad2)
        return net_grad, leaves_grad.t()


class Route(torch.autograd.Function):
    @staticmethod
    def forward(ctx: Any, *args: Any, **kwargs: Any) -> Tensor:
        (
            net,
            tree_arange,
            batch_indices,
            ones,
            increment_masks,
            num_tree,
            num_batch,
            tree_depth,
            num_internals,
        ) = args
        shape = num_batch, -1, num_internals
        sigmoid_net = torch.sigmoid(net)
        p_left = sigmoid_net.view(*shape).transpose(0, 1)
        p_right = 1.0 - p_left
        flat_probabilities = torch.cat([p_left, p_right], dim=-1)
        flat_probabilities = flat_probabilities.contiguous().view(num_tree, -1)
        current_indices = batch_indices + increment_masks[0]
        flat_dim = flat_probabilities.shape[-1]
        tree_arange = tree_arange * flat_dim
        routes = flat_probabilities.take(tree_arange + current_indices[None, ...])
        all_routes = [routes.clone()]
        for i in range(1, tree_depth + 1):
            current_indices = batch_indices + increment_masks[i]
            current_indices = tree_arange + current_indices[None, ...]
            current_routes = flat_probabilities.take(current_indices)
            all_routes.append(current_routes)
            routes *= current_routes
        ctx.save_for_backward(ones, sigmoid_net, *all_routes)
        ctx.tree_depth = tree_depth
        ctx.num_tree = num_tree
        return routes

    @staticmethod
    def backward(ctx: Any, *grad_outputs: Any) -> Tuple[Optional[Tensor], ...]:
        grad_output = grad_outputs[0]
        dummy_grads = tuple(None for _ in range(8))
        if grad_output is None:
            return (None,) + dummy_grads
        num_tree = ctx.num_tree
        tree_depth = ctx.tree_depth
        ones_list, sigmoid_net, *all_routes = ctx.saved_tensors
        cursor = 0
        divide = 1
        num_leaves = 2 ** (tree_depth + 1)
        sub_grads_shape = num_tree, sigmoid_net.shape[0], num_leaves - 1
        sub_grads = torch.zeros(*sub_grads_shape, device=grad_output.device)
        for i in range(tree_depth + 1):
            ones = ones_list[i]
            nodes = ones[None, None, ...]
            for j in range(tree_depth + 1):
                if j == i:
                    continue
                nodes = nodes * all_routes[j]
            sub_grad = grad_output * nodes
            section = int(round(num_leaves / divide))
            sub_grad = sub_grad.view(num_tree, -1, divide, section)
            sub_grad = sub_grad.sum(-1)
            sub_grads[..., cursor : cursor + divide] = sub_grad
            cursor += divide
            divide *= 2

        sub_grads = sub_grads.transpose(0, 1).contiguous()
        sub_grads = sub_grads.view(-1, num_tree * (num_leaves - 1))
        return (sigmoid_net * (1.0 - sigmoid_net) * sub_grads,) + dummy_grads


class DNDF(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        num_tree: int = 10,
        tree_depth: int = 4,
        is_regression: Optional[bool] = None,
        tree_proj_config: Optional[Dict[str, Any]] = None,
        use_fast_dndf: bool = True,
    ):
        super().__init__()
        self._num_tree = num_tree
        self._tree_depth = tree_depth
        self._is_regression = out_dim == 1 if is_regression is None else is_regression
        self._num_leaf = 2 ** (self._tree_depth + 1)
        self._num_internals = self._num_leaf - 1
        self._output_dim = out_dim
        self._fast = use_fast_dndf
        if tree_proj_config is None:
            tree_proj_config = {}
        tree_proj_config.setdefault("pruner_config", {})
        self.tree_proj = Linear(
            in_dim,
            self._num_internals * self._num_tree,
            **tree_proj_config,
        )
        leaves_shape = self._num_tree * self._num_leaf, self._output_dim
        self.leaves = nn.Parameter(torch.empty(*leaves_shape))
        with torch.no_grad():
            torch.nn.init.xavier_uniform_(self.leaves.data)
        # buffers
        num_repeat, num_local_internals = self._num_leaf // 2, 1
        ones_np = np.repeat([1, -1], num_repeat)
        ones_list = [torch.from_numpy(ones_np.astype(np_float_type))]
        increment_indices_np = np.repeat([0, self._num_internals], num_repeat)
        increment_indices = [torch.from_numpy(increment_indices_np.astype(np_int_type))]
        for i in range(1, self._tree_depth + 1):
            num_repeat //= 2
            num_local_internals *= 2
            arange = np.arange(num_local_internals - 1, 2 * num_local_internals - 1)
            ones_np = np.repeat([1, -1], num_repeat)
            ones_np = np.tile(ones_np, 2 ** i)
            ones_list.append(torch.from_numpy(ones_np.astype(np_float_type)))
            increment_mask = np.repeat(arange, 2)
            increment_mask += np.tile([0, self._num_internals], num_local_internals)
            increment_mask = np.repeat(increment_mask, num_repeat)
            increment_mask = torch.from_numpy(increment_mask.astype(np_int_type))
            increment_indices.append(increment_mask)
        self.increment_masks: Tensor
        self.register_buffer("tree_arange", torch.arange(num_tree)[..., None, None])
        self.register_buffer("ones", torch.stack(ones_list))
        self.register_buffer("increment_indices", torch.stack(increment_indices))

    def forward(self, net: Tensor) -> Tensor:
        num_batch = net.shape[0]
        tree_net = self.tree_proj(net)

        num_flat_prob = 2 * self._num_internals
        arange_args = 0, num_flat_prob * num_batch, num_flat_prob
        batch_indices = torch.arange(*arange_args, device=tree_net.device).view(-1, 1)

        if self._fast:
            routes = Route.apply(
                tree_net,
                self.tree_arange,
                batch_indices,
                self.ones,
                self.increment_indices,
                self._num_tree,
                num_batch,
                self._tree_depth,
                self._num_internals,
            )
        else:
            shape = num_batch, -1, self._num_internals
            p_left = torch.sigmoid(tree_net).view(*shape).transpose(0, 1)
            p_right = 1.0 - p_left
            flat_probabilities = torch.cat([p_left, p_right], dim=-1).contiguous()
            flat_probabilities = flat_probabilities.view(self._num_tree, -1)
            current_indices = batch_indices + self.increment_indices[0]  # type: ignore
            flat_dim = flat_probabilities.shape[-1]
            tree_arange = self.tree_arange * flat_dim  # type: ignore
            routes = flat_probabilities.take(tree_arange + current_indices[None, ...])
            for i in range(1, self._tree_depth + 1):
                current_indices = batch_indices + self.increment_indices[i]  # type: ignore
                current_indices = tree_arange + current_indices[None, ...]
                routes *= flat_probabilities.take(current_indices)

        features = routes.transpose(0, 1).contiguous().view(num_batch, -1)
        if self._is_regression or self._output_dim <= 1:
            outputs = features.mm(self.leaves)
        else:
            if self._fast:
                outputs = LeafAggregation.apply(features, self.leaves)
            else:
                leaves = F.softmax(self.leaves, dim=1)
                outputs = features.mm(leaves)
        return outputs / self._num_tree


class CrossBase(Module, metaclass=ABCMeta):
    @abstractmethod
    def forward(self, x: Tensor, x0: Tensor) -> Tensor:
        pass


class InnerCross(CrossBase):
    def __init__(self, dim: int, **kwargs: Any):
        super().__init__()
        self.inner = Linear(dim, 1, bias=False, **kwargs)

    def forward(self, x: Tensor, x0: Tensor) -> Tensor:
        return x0 * self.inner(x)


class CrossBlock(Module):
    def __init__(
        self,
        dim: int,
        bias: bool = True,
        residual: bool = True,
        *,
        cross_builder: Callable[[int], Module] = None,
        **kwargs: Any,
    ):
        super().__init__()
        if cross_builder is None:
            cross_builder = lambda dim_: InnerCross(dim_)
        self.cross = cross_builder(dim)
        if not bias:
            self.bias = None
        else:
            self.bias = nn.Parameter(torch.empty(1, dim))
            with torch.no_grad():
                self.bias.data.fill_(kwargs.get("bias_fill", 0.0))
        self.residual = residual

    def forward(self, x: Tensor, x0: Tensor) -> Tensor:
        crossed = self.cross(x, x0)
        if self.residual:
            crossed = crossed + x
        if self.bias is None:
            return crossed
        return crossed + self.bias

    def extra_repr(self) -> str:
        return f"(bias): {False if self.bias is None else True}"


class TreeResBlock(Module):
    def __init__(self, dim: int, dndf_config: Optional[Dict[str, Any]] = None):
        super().__init__()
        if dndf_config is None:
            dndf_config = {}
        self.dim = float(dim)
        self.in_dndf = DNDF(dim, dim, **shallow_copy_dict(dndf_config))
        self.inner_dndf = DNDF(dim, dim, **shallow_copy_dict(dndf_config))

    def forward(self, net: Tensor) -> Tensor:
        res = self.in_dndf(net)
        res = self.dim * res - 1.0
        res = self.inner_dndf(res)
        res = self.dim * res - 1.0
        return net + res


class InvertibleBlock(Module):
    def __init__(
        self,
        dim: int,
        *,
        default_activation: str = "mish",
        transition_builder: Callable[[int], Module] = None,
    ):
        if dim % 2 != 0:
            raise ValueError("`dim` should be divided by 2")
        super().__init__()
        h_dim = int(dim // 2)
        # transition
        if transition_builder is not None:
            transition = transition_builder(dim)
        else:
            transition = MLP.simple(
                h_dim,
                None,
                [h_dim],
                activation=default_activation,
            )
        self.transition = transition

    def forward(self, net1: Tensor, net2: Tensor) -> tensor_tuple_type:
        net1 = net1 + self.transition(net2)
        return net2, net1

    def inverse(self, net1: Tensor, net2: Tensor) -> tensor_tuple_type:
        net2 = net2 - self.transition(net1)
        return net2, net1


class ConditionalOutput(NamedTuple):
    net: Tensor
    cond: Tensor
    responses: List[Tensor]


class PseudoInvertibleBlock(Module):
    def __init__(
        self,
        in_dim: int,
        latent_dim: int,
        out_dim: int,
        *,
        to_activation: str = "mish",
        from_activation: str = "mish",
        to_transition_builder: Optional[Callable[[int, int], Module]] = None,
        from_transition_builder: Optional[Callable[[int, int], Module]] = None,
    ):
        super().__init__()
        if to_transition_builder is not None:
            self.to_latent = to_transition_builder(in_dim, latent_dim)
        else:
            num_units = [latent_dim, latent_dim, latent_dim]
            self.to_latent = MLP.simple(
                in_dim,
                None,
                num_units,
                activation=to_activation,
            )
        if from_transition_builder is not None:
            self.from_latent = from_transition_builder(latent_dim, out_dim)
        else:
            self.from_latent = MLP.simple(
                latent_dim,
                out_dim,
                [latent_dim, latent_dim],
                bias=True,
                activation=from_activation,
            )

    def forward(
        self,
        net: Union[Tensor, Any],
        cond: Optional[Union[Tensor, Any]] = None,
    ) -> Union[Tensor, Any]:
        if cond is None:
            return self.to_latent(net)
        return self.to_latent(net, cond)

    def inverse(
        self,
        net: Union[Tensor, Any],
        cond: Optional[Union[Tensor, Any]] = None,
    ) -> Union[Tensor, Any]:
        if cond is None:
            return self.from_latent(net)
        return self.from_latent(net, cond)


class MonotonousMapping(Module):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        ascent: bool,
        bias: bool = True,
        dropout: float = 0.0,
        batch_norm: bool = False,
        activation: Optional[str] = None,
        init_method: Optional[str] = "xavier_uniform",
        positive_transform: str = "square",
        scaler: Optional[float] = None,
        use_scaler: bool = True,
        **kwargs: Any,
    ):
        super().__init__()
        self.ascent = ascent
        self.positive_transform = positive_transform
        self.config = shallow_copy_dict(kwargs)
        # linear
        self.linear = Linear(
            in_dim,
            out_dim,
            bias=bias,
            init_method=init_method,
            **kwargs,
        )
        # dropout
        use_dropout = 0.0 < dropout < 1.0
        self.dropout = None if not use_dropout else Dropout(dropout)
        # batch norm
        self.bn = None if not batch_norm else BN(out_dim)
        # activation
        if activation is None:
            self.activation: Optional[Module] = None
        else:
            activation_config = self.config.setdefault("activation_config", None)
            self.activation = Activations.make(activation, activation_config)
        # scaler
        if not use_scaler:
            self.scaler = None
        else:
            if scaler is None:
                scaler = math.log(math.e - 1)
            self.scaler = nn.Parameter(torch.full([out_dim, 1], scaler))

    def _get_positive_weight(self) -> Tensor:
        weight = self.linear.weight
        if self.positive_transform == "abs":
            pos_weight = torch.abs(weight)
        elif self.positive_transform == "square":
            pos_weight = weight ** 2
        elif self.positive_transform == "sigmoid":
            pos_weight = torch.sigmoid(weight)
        elif self.positive_transform == "softmax":
            pos_weight = F.softmax(weight, dim=1)
        elif self.positive_transform == "softplus":
            pos_weight = F.softplus(weight)
        else:
            msg = f"positive transform '{self.positive_transform}' is not implemented"
            raise NotImplementedError(msg)
        if self.scaler is None:
            return pos_weight
        scaler = F.softplus(self.scaler)
        return scaler * pos_weight

    def forward(self, net: Tensor, *, reuse: bool = False) -> Tensor:
        weight = self._get_positive_weight()
        if not self.ascent:
            weight = -weight
        net = F.linear(net, weight, self.linear.bias)
        if self.bn is not None:
            net = self.bn(net)
        if self.activation is not None:
            net = self.activation(net)
        if self.dropout is not None:
            net = self.dropout(net, reuse=reuse)
        return net

    def extra_repr(self) -> str:
        msg = f"(ascent): {self.ascent}\n(positive): {self.positive_transform}"
        if self.scaler is None:
            return msg
        return f"{msg}\n(scaler): {self.scaler.shape}"

    @classmethod
    def make_couple(
        cls,
        in_dim: int,
        hidden_dim: int,
        out_dim: int,
        activation: str,
        *,
        ascent: bool,
        bias: bool = True,
        dropout: float = 0.0,
        batch_norm: bool = False,
        use_residual: bool = False,
        use_inverse_activation: bool = True,
        scaler: Optional[float] = None,
        init_method: Optional[str] = "normal",
        **kwargs: Any,
    ) -> Module:
        if not use_inverse_activation:
            inverse_activation = None
        elif activation == "tanh":
            inverse_activation = "atanh"
        elif activation == "sigmoid":
            inverse_activation = "logit"
        elif activation == "softplus":
            inverse_activation = "isoftplus"
        else:
            msg = f"inverse activation of '{activation}' is not defined"
            raise NotImplementedError(msg)
        initialize_config = kwargs.setdefault("initialize_config", {})
        initialize_config.setdefault("std", 3.0)
        squash_unit = cls(
            in_dim,
            hidden_dim,
            ascent=ascent,
            bias=bias,
            dropout=dropout,
            batch_norm=batch_norm,
            activation=activation,
            init_method=init_method,
            positive_transform="softmax" if in_dim > 1 else "softplus",
            scaler=scaler,
            **kwargs,
        )
        inverse_unit = cls(
            hidden_dim,
            out_dim,
            ascent=True,
            bias=False,
            dropout=dropout,
            batch_norm=batch_norm,
            activation=inverse_activation,
            init_method=init_method,
            positive_transform="softmax",
            use_scaler=False,
            **kwargs,
        )

        class Couple(Module):
            def __init__(self) -> None:
                super().__init__()
                self.squash = squash_unit
                self.inverse = inverse_unit
                self.use_residual = use_residual

            def forward(self, net: Tensor) -> Tensor:
                squashed = self.squash(net)
                inversed = self.inverse(squashed)
                if not self.use_residual:
                    return inversed
                if in_dim == out_dim:
                    return net + inversed
                if hidden_dim == out_dim:
                    return squashed + inversed
                raise ValueError(
                    "either `in_dim` or `hidden_dim` should be identical with "
                    "`out_dim` when `use_residual` is True"
                )

            def extra_repr(self) -> str:
                return f"(residual): {self.use_residual}"

        return Couple()

    @classmethod
    def stack(
        cls,
        in_dim: int,
        out_dim: Optional[int],
        num_units: List[int],
        *,
        ascent: bool,
        bias: bool = False,
        dropout: float = 0.0,
        batch_norm: bool = False,
        final_batch_norm: bool = False,
        use_couple: bool = True,
        use_couple_bias: bool = True,
        use_couple_residual: bool = False,
        use_couple_inverse_activation: bool = True,
        activation: Optional[str] = "sigmoid",
        init_method: Optional[str] = "normal",
        positive_transform: str = "softmax",
        use_scaler: bool = True,
        scaler: Optional[float] = None,
        return_blocks: bool = False,
        **kwargs: Any,
    ) -> Union[List[Module], Module]:
        blocks = []
        common_kwargs = {
            "ascent": ascent,
            "dropout": dropout,
            "batch_norm": batch_norm,
            "activation": activation,
        }

        def _make(
            in_dim_: int,
            out_dim_: int,
            hidden_dim: int,
            use_residual: bool,
            use_inverse_activation: bool,
        ) -> Module:
            local_kwargs = shallow_copy_dict(common_kwargs)
            local_kwargs.update(shallow_copy_dict(kwargs))
            local_kwargs["in_dim"] = in_dim_
            local_kwargs["out_dim"] = out_dim_
            if use_couple:
                local_kwargs["scaler"] = scaler
                local_kwargs["hidden_dim"] = hidden_dim
                local_kwargs["use_residual"] = use_residual
                local_kwargs["use_inverse_activation"] = use_inverse_activation
                local_kwargs["bias"] = use_couple_bias
                return cls.make_couple(**local_kwargs)
            local_kwargs["bias"] = bias
            local_kwargs["init_method"] = init_method
            local_kwargs["positive_transform"] = positive_transform
            local_kwargs["use_scaler"] = use_scaler
            return cls(**local_kwargs)

        current_in_dim = in_dim
        for num_unit in num_units:
            blocks.append(
                _make(
                    current_in_dim,
                    num_unit,
                    num_unit,
                    use_couple_residual,
                    True,
                )
            )
            common_kwargs["ascent"] = True
            current_in_dim = num_unit
        if out_dim is not None:
            common_kwargs["batch_norm"] = final_batch_norm
            blocks.append(
                _make(
                    current_in_dim,
                    out_dim,
                    current_in_dim,
                    False,
                    use_couple_inverse_activation,
                )
            )

        if return_blocks:
            return blocks
        return nn.Sequential(*blocks)


class ConditionalBlocks(Module):
    def __init__(
        self,
        main_blocks: ModuleList,
        condition_blocks: ModuleList,
        detach_condition: bool = False,
        *,
        add_last: bool,
        cond_mixtures: Optional[ModuleList] = None,
    ):
        super().__init__()
        self.add_last = add_last
        self.detach_condition = detach_condition
        self.num_blocks = len(main_blocks)
        if self.num_blocks != len(condition_blocks):
            msg = "`main_blocks` and `condition_blocks` should have same sizes"
            raise ValueError(msg)
        self.main_blocks = main_blocks
        self.condition_blocks = condition_blocks
        self.cond_mixtures = cond_mixtures

    def forward(
        self,
        net: Tensor,
        cond: Union[Tensor, List[Tensor]],
    ) -> ConditionalOutput:
        if isinstance(cond, list):
            cond_responses = cond
            detached_responses = [net.detach() for net in cond]
        else:
            cond_responses = []
            detached_responses = []
            for condition in self.condition_blocks:
                cond = condition(cond)
                cond_responses.append(cond)  # type: ignore
                detached_responses.append(cond.detach())  # type: ignore
        responses = detached_responses if self.detach_condition else cond_responses
        iterator = enumerate(zip(self.main_blocks, responses))
        for i, (main, response) in iterator:
            net = main(net)
            if i < self.num_blocks - 1 or self.add_last:
                if self.cond_mixtures is None:
                    net = net + response
                else:
                    net = self.cond_mixtures[i](net, response)
        return ConditionalOutput(net, cond_responses[-1], responses)

    def extra_repr(self) -> str:
        add_last = f"(add_last): {self.add_last}"
        detach_condition = f"(detach_condition): {self.detach_condition}"
        return f"{add_last}\n{detach_condition}"


class AttentionOutput(NamedTuple):
    output: Tensor
    weights: Tensor


class Attention(Module):
    def __init__(
        self,
        input_dim: int,
        num_heads: int = 1,
        *,
        dropout: float = 0.0,
        is_self_attention: bool = False,
        k_dim: Optional[int] = None,
        v_dim: Optional[int] = None,
        embed_dim: Optional[int] = None,
        activation: Optional[str] = None,
        activation_config: Optional[Dict[str, Any]] = None,
        q_linear_config: Optional[Dict[str, Any]] = None,
        k_linear_config: Optional[Dict[str, Any]] = None,
        v_linear_config: Optional[Dict[str, Any]] = None,
        in_linear_config: Optional[Dict[str, Any]] = None,
        out_linear_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self.input_dim = input_dim
        self.is_self_attn = is_self_attention
        if not is_self_attention:
            self.k_dim = k_dim if k_dim is not None else input_dim
            self.v_dim = v_dim if v_dim is not None else input_dim
        else:
            if k_dim is not None and k_dim != input_dim:
                raise ValueError("self attention is used but `k_dim` != `input_dim`")
            if v_dim is not None and v_dim != input_dim:
                raise ValueError("self attention is used but `v_dim` != `input_dim`")
            self.k_dim = self.v_dim = input_dim
        if embed_dim is None:
            embed_dim = min(32, input_dim) * num_heads
        self.embed_dim = embed_dim

        self.num_heads = num_heads
        self.head_dim = self.embed_dim // num_heads
        self.scaling = float(self.head_dim) ** -0.5
        if self.head_dim * num_heads != self.embed_dim:
            raise ValueError("`embed_dim` must be divisible by `num_heads`")

        def _warn(prefix: str) -> None:
            msg = (
                f"self attention is used so `{prefix}_linear_config` will be ignored, "
                "please use `in_linear_config` instead"
            )
            print(f"{LoggingMixin.warning_prefix}{msg}")

        if q_linear_config is None:
            q_linear_config = {}
        elif is_self_attention:
            _warn("q")
        if k_linear_config is None:
            k_linear_config = {}
        elif is_self_attention:
            _warn("k")
        if v_linear_config is None:
            v_linear_config = {}
        elif is_self_attention:
            _warn("v")

        if is_self_attention:
            if in_linear_config is None:
                in_linear_config = {}
            self.in_linear = Linear(input_dim, 3 * self.embed_dim, **in_linear_config)
        else:
            self.q_linear = Linear(input_dim, self.embed_dim, **q_linear_config)
            self.k_linear = Linear(self.k_dim, self.embed_dim, **k_linear_config)
            self.v_linear = Linear(self.v_dim, self.embed_dim, **v_linear_config)

        if out_linear_config is None:
            out_linear_config = {}
        self.out_linear = Linear(self.embed_dim, input_dim, **out_linear_config)

        self.dropout = dropout
        self.activation = Activations.make(activation, activation_config)

    def _to_heads(self, tensor: Tensor) -> Tensor:
        batch_size, seq_len, in_feature = tensor.shape
        tensor = tensor.view(batch_size, seq_len, self.num_heads, self.head_dim)
        return tensor.permute(0, 2, 1, 3).contiguous().view(-1, seq_len, self.head_dim)

    def forward(
        self,
        q: Tensor,
        k: Tensor,
        v: Tensor,
        mask: Optional[Tensor] = None,
    ) -> AttentionOutput:
        # `mask` represents slots which will be zeroed
        k_len = k.shape[1]
        if self.is_self_attn:
            q, k, v = self.in_linear(q).chunk(3, dim=-1)
        else:
            # B, Sq, Din -> B, Sq, D
            q = self.q_linear(q)
            # B, Sk, Dk -> B, Sk, D
            k = self.k_linear(k)
            # B, Sv, Dv -> B, Sk, D
            v = self.v_linear(v)
        q, k, v = map(self.activation, [q, k, v])
        # scale
        q = q * self.scaling
        # B, S*, D -> B * N_head, S*, D_head
        q, k, v = map(self._to_heads, [q, k, v])
        if mask is not None:
            # B, Sq, Sk -> B * N_head, Sq, Sk
            mask = mask.repeat(self.num_heads, 1, 1)
        # B * N_head, Sq, Sk
        raw_weights = torch.bmm(q, k.transpose(-2, -1))
        if mask is not None:
            raw_weights.masked_fill_(mask, float("-inf"))
        # B * N_head, Sq, Sk -> # B * N_head, Sq, Sk
        weights = F.softmax(raw_weights, dim=-1)
        if 0.0 < self.dropout < 1.0:
            weights = F.dropout(weights, self.dropout, self.training)
        # B * N_head, Sq, D_head
        output = torch.bmm(weights, v)
        # B * N_head, Sq, D_head -> B, N_head, Sq, D_head
        nb, q_len, d_head = output.shape
        output = output.view(nb // self.num_heads, self.num_heads, q_len, d_head)
        # B, N_head, Sq, D_head -> B, Sq, D
        output = output.permute(0, 2, 1, 3).contiguous()
        output = output.view(-1, q_len, self.embed_dim)
        # B, Sq, D -> B, Sq, Din
        output = self.activation(self.out_linear(output))
        return AttentionOutput(output, weights.view(-1, self.num_heads, q_len, k_len))


__all__ = [
    "Linear",
    "Mapping",
    "MLP",
    "DNDF",
    "CrossBlock",
    "TreeResBlock",
    "InvertibleBlock",
    "PseudoInvertibleBlock",
    "MonotonousMapping",
    "ConditionalBlocks",
    "ConditionalOutput",
    "Attention",
]



import torch

from typing import Any
from typing import Dict
from typing import List
from typing import Union
from typing import Optional

from ..base import HeadBase
from ...blocks import BN
from ...blocks import MLP


@HeadBase.register("fcnn")
class FCNNHead(HeadBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        hidden_units: List[int],
        mapping_configs: Optional[Union[Dict[str, Any], List[Dict[str, Any]]]] = None,
        final_mapping_config: Optional[Dict[str, Any]] = None,
        *,
        use_final_bn: bool = False,
        **kwargs: Any,
    ):
        super().__init__(in_dim, out_dim, **kwargs)
        if mapping_configs is None:
            mapping_configs = {}
        self.mlp = MLP(
            in_dim,
            out_dim,
            hidden_units,
            mapping_configs,
            final_mapping_config=final_mapping_config,
        )
        if not use_final_bn:
            self.final_bn = None
        else:
            self.final_bn = BN(out_dim)

    def forward(self, net: torch.Tensor) -> torch.Tensor:
        net = self.mlp(net)
        if self.final_bn is not None:
            net = self.final_bn(net)
        return net


__all__ = ["FCNNHead"]



import torch

import numpy as np

from typing import Any
from typing import Dict
from typing import Union
from typing import Optional
from cftool.ml import Metrics
from cftool.misc import is_numeric
from cftool.misc import timing_context

from ...misc.toolkit import *
from ..base import ModelBase
from ...types import tensor_dict_type
from ...protocol import TrainerState
from ...modules.transform.core import SplitFeatures


@ModelBase.register("ddr")
@ModelBase.register_pipe("ddr")
class DDR(ModelBase):
    def _init_config(self) -> None:
        super()._init_config()
        # pipe
        head_config = self.get_pipe_config("ddr", "head")
        self.fetch_cdf = head_config.setdefault("fetch_cdf", True)
        self.fetch_q = head_config.setdefault("fetch_q", True)
        if not self.fetch_q and not self.fetch_cdf:
            raise ValueError("something must be fetched, either `q` or `cdf`")
        # common
        self.q_metric = Metrics("quantile")
        self.config.setdefault("ema_decay", 0.0)
        self._synthetic_step = self.config.setdefault("synthetic_step", 10)
        self._synthetic_range = self.config.setdefault("synthetic_range", 3.0)
        labels = self.tr_data.processed.y
        self.y_min, self.y_max = labels.min(), labels.max()
        self.y_diff = self.y_max - self.y_min
        quantile_anchors = np.linspace(0.005, 0.995, 100).astype(np.float32)[..., None]
        y_anchor_choices = quantile_anchors * self.y_diff + self.y_min
        self.register_buffer("quantile_anchors", torch.from_numpy(quantile_anchors))
        self.register_buffer("y_anchor_choices", torch.from_numpy(y_anchor_choices))
        # loss config
        self.config["loss"] = "ddr"
        loss_config = self.config.setdefault("loss_config", {})
        loss_config.setdefault("mtl_method", None)
        loss_config["fetch_q"] = self.fetch_q
        loss_config["fetch_cdf"] = self.fetch_cdf
        # trainer config
        default_metric_types = ["loss"]
        self.quantile_metric_config = None
        if self.fetch_q:
            default_metric_types += ["ddr", "quantile"]
            default_q = np.linspace(0.1, 0.9, 5).tolist()
            qmq = self.config.setdefault("quantile_metric_q", default_q)
            qmq = np.asarray(qmq, np.float32)
            self.quantile_metric_config = {"q": qmq}
        if self.fetch_cdf:
            default_metric_types += ["cdf", "pdf"]
        if self.fetch_q and self.fetch_cdf:
            default_metric_types += ["q_recover", "y_recover"]
        default_metric_weights = {
            "loss": 1.0,
            "ddr": 5.0,
            "quantile": 1.0,
            "cdf": 5.0,
            "pdf": 1.0,
            "q_recover": 10.0,
            "y_recover": 1.0,
        }
        # pipeline config
        # TODO : Support ONNX in DDR
        production = "pipeline"
        # inject new default
        new_default_config = {
            "production": production,
            "trainer_config": {
                "clip_norm": 1.0,
                "num_epoch": 40,
                "max_epoch": 1000,
                "metric_config": {
                    "decay": 0.0,
                    "types": default_metric_types,
                    "weights": default_metric_weights,
                    "quantile_config": self.quantile_metric_config,
                },
            },
        }
        self.environment.update_default_config(new_default_config)

    # utilities

    def _expand(
        self,
        n: int,
        elem: Union[float, torch.Tensor],
        *,
        numpy: bool = False,
        to_device: bool = True,
    ) -> Union[torch.Tensor, np.ndarray]:
        if isinstance(elem, torch.Tensor):
            return elem
        if not is_numeric(elem):
            elem_arr = np.asarray(elem, np.float32)
        else:
            elem_arr = np.repeat(elem, n).astype(np.float32)
        elem_arr = elem_arr.reshape([-1, 1])
        if numpy:
            return elem_arr
        elem_tensor = torch.from_numpy(elem_arr)
        if to_device:
            elem_tensor = elem_tensor.to(self.device)
        return elem_tensor

    # core

    def _quantile(
        self,
        net: Union[torch.Tensor, SplitFeatures],
        q_batch: Optional[torch.Tensor],
        do_inverse: bool,
    ) -> tensor_dict_type:
        kwargs = {
            "ddr": {
                "q_batch": q_batch,
                "median": q_batch is None,
                "do_inverse": do_inverse,
            }
        }
        results = self.execute(net, clear_cache=False, head_kwargs_dict=kwargs)
        ddr_results = results["ddr"]
        return ddr_results  # type: ignore

    def _median(self, net: Union[torch.Tensor, SplitFeatures]) -> tensor_dict_type:
        results = self.execute(
            net,
            clear_cache=False,
            head_kwargs_dict={"ddr": {"median": True}},
        )
        ddr_results = results["ddr"]
        return ddr_results  # type: ignore

    def _cdf(
        self,
        net: Union[torch.Tensor, SplitFeatures],
        y_batch: torch.Tensor,
        need_optimize: bool,
        return_pdf: bool,
        do_inverse: bool,
    ) -> tensor_dict_type:
        use_grad = self.training or return_pdf
        y_batch.requires_grad_(return_pdf)
        with mode_context(self, to_train=None, use_grad=use_grad):
            kwargs = {"ddr": {"y_batch": y_batch, "do_inverse": do_inverse}}
            results = self.execute(net, clear_cache=False, head_kwargs_dict=kwargs)
        ddr_results = results["ddr"]  # type: ignore
        cdf = ddr_results["q"]  # type: ignore
        if not return_pdf:
            pdf = None
        else:
            pdf = get_gradient(cdf, y_batch, need_optimize, need_optimize)
            y_batch.requires_grad_(False)
        return {
            "cdf": cdf,
            "pdf": pdf,
            "cdf_logit": ddr_results["q_logit"],  # type: ignore
            "cdf_logit_mul": ddr_results["q_logit_mul"],  # type: ignore
            "y_inverse_res": ddr_results["y_inverse_res"],  # type: ignore
        }

    def _core(
        self,
        batch_size: int,
        split: SplitFeatures,
        y_batch: torch.Tensor,
        synthetic: bool,
    ) -> tensor_dict_type:
        mask: Optional[torch.Tensor] = None
        if synthetic:
            random_indices = torch.randint(2, [batch_size, 1], device=y_batch.device)
            mask = random_indices == 0
        # TODO : try to optimize this when `fetches` is not full
        # generate quantile / anchor batch
        with timing_context(self, "forward.generate_anchors"):
            q_synthetic_batch = None
            if self.training:
                q_batch = torch.empty_like(y_batch).uniform_()
                y_batch = torch.empty_like(y_batch).uniform_()
                y_batch = y_batch * self.y_diff + self.y_min
                # y_batch = y_batch[np.random.permutation(batch_size)].detach()
                if synthetic:
                    assert mask is not None
                    q_synthetic_batch = torch.empty_like(y_batch)
                    q_synthetic_batch[mask] = 0.25
                    q_synthetic_batch[~mask] = 0.75
            else:
                q_batch = self.quantile_anchors  # type: ignore
                y_batch = self.y_anchor_choices  # type: ignore
                n_repeat = int(batch_size / len(q_batch)) + 1
                q_batch = q_batch.repeat_interleave(n_repeat, dim=0)[:batch_size]
                y_batch = y_batch.repeat_interleave(n_repeat, dim=0)[:batch_size]
        # build predictions
        with timing_context(self, "forward.median"):
            median_rs = {}
            rs = self._median(split)
            if not synthetic:
                median_rs = {
                    "predictions": rs["median"],
                    "med_pos_med_res": rs["pos_med_res"],
                    "med_neg_med_res": rs["neg_med_res"],
                }
        # synthetic predictions
        with timing_context(self, "forward.synthetic"):
            if synthetic:
                if self.fetch_q:
                    assert q_synthetic_batch is not None
                    q_rs = self._quantile(split, q_synthetic_batch, True)
                    median_rs["syn_med_mul"] = q_rs["med_mul"]
                    if not self.fetch_cdf:
                        return median_rs
                if self.fetch_cdf:
                    median = rs["median"].detach()
                    pos_med_res = rs["pos_med_res"]
                    neg_med_res = rs["neg_med_res"]
                    assert mask is not None
                    y_syn_batch = torch.where(mask, pos_med_res, -neg_med_res)
                    y_syn_batch = y_syn_batch.detach() + median
                    y_rs = self._cdf(split, y_syn_batch, False, False, False)
                    y_med_rs = self._cdf(split, median, False, False, False)
                    median_rs.update(
                        {
                            "syn_cdf_logit_mul": y_rs["cdf_logit_mul"],
                            "syn_med_cdf_logit_mul": y_med_rs["cdf_logit_mul"],
                        }
                    )
                    if not self.fetch_q:
                        return median_rs
        # TODO : Some of the calculations in `forward.median` could be reused
        with timing_context(self, "forward.quantile"):
            if not self.fetch_q:
                q_rs = {}
            else:
                q_rs = self._quantile(split, q_batch, True)
        with timing_context(self, "forward.cdf"):
            if not self.fetch_cdf:
                y_rs = {}
            else:
                cdf_inverse = not synthetic
                y_rs = self._cdf(split, y_batch, True, True, cdf_inverse)
        # construct results
        net = list(self._transform_cache.values())[0]
        results: tensor_dict_type = {"net": net, "q_batch": q_batch, "y_batch": y_batch}
        results.update({k: v for k, v in median_rs.items() if v is not None})
        results.update({k: v for k, v in q_rs.items() if v is not None})
        results.update({k: v for k, v in y_rs.items() if v is not None})
        for key in set(median_rs.keys()) | set(q_rs.keys()) | set(y_rs.keys()):
            results.setdefault(key, None)
        return results

    def _predict_quantile(
        self,
        split: SplitFeatures,
        batch_size: int,
        kwargs: Dict[str, Any],
        q: Optional[np.ndarray] = None,
    ) -> Dict[str, torch.Tensor]:
        if not self.fetch_q:
            raise ValueError("quantile function is not fetched")
        if q is not None:
            q = q.tolist()
        else:
            q = kwargs.get("q")
        if q is None:
            raise ValueError(f"quantile cannot be predicted without q")
        quantiles_list, med_mul_list = [], []
        q_list = [q] if isinstance(q, float) else q
        for q in q_list:
            q_batch = self._expand(batch_size, q)
            pack = self._quantile(split, q_batch, False)
            quantiles_list.append(pack["median"] + pack["y_res"])
            med_mul_list.append(pack["med_mul"])
        if len(q_list) == 1:
            return {"quantiles": quantiles_list[0], "med_mul": med_mul_list[0]}
        return {
            "quantiles": torch.cat(quantiles_list, dim=1),
            "med_mul": torch.cat(med_mul_list, dim=1),
        }

    # API

    def forward(
        self,
        batch: tensor_dict_type,
        batch_idx: Optional[int] = None,
        state: Optional[TrainerState] = None,
        batch_indices: Optional[np.ndarray] = None,
        loader_name: Optional[str] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        # pre-processing
        x_batch = batch["x_batch"]
        labels = batch[self.labels_key]
        batch_size = x_batch.shape[0]
        split = self._split_features(x_batch, batch_indices, loader_name)
        forward_dict = {}
        # check median residual inference
        predict_mr = kwargs.get("predict_median_residual", False)
        if predict_mr:
            pack = self._median(split)
            median = pack["median"]
            pos, neg = pack["pos_med_res"], pack["neg_med_res"]
            forward_dict["mr_pos"] = pos + median
            forward_dict["mr_neg"] = -neg + median
        # check q inference
        predict_quantile = kwargs.get("predict_quantiles")
        if predict_quantile:
            forward_dict.update(self._predict_quantile(split, batch_size, kwargs))
        # check y inference
        predict_pdf = kwargs.get("predict_pdf", False)
        predict_cdf = kwargs.get("predict_cdf", False)
        if predict_pdf or predict_cdf:
            if not self.fetch_cdf:
                raise ValueError("cdf function is not fetched")
            y = kwargs.get("y")
            if y is None:
                raise ValueError(f"pdf / cdf cannot be predicted without y")
            labels = self._expand(batch_size, y, numpy=True)
            labels = self.tr_data.transform_labels(labels)
            labels = to_torch(labels).to(self.device)
            forward_dict = self._cdf(split, labels, False, predict_pdf, False)
        # check quantile metric
        getting_metrics = kwargs.get("getting_metrics", False)
        if getting_metrics and self.quantile_metric_config is not None:
            q = self.quantile_metric_config["q"]
            forward_dict.update(self._predict_quantile(split, batch_size, kwargs, q))
        # forward
        return_loss = kwargs.get("return_loss", False)
        if return_loss or not forward_dict:
            forward_dict.update(self._core(batch_size, split, labels, False))
        self.clear_execute_cache()
        return forward_dict

    def loss_function(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        batch_indices: Optional[torch.Tensor],
        forward_results: tensor_dict_type,
        state: Optional[TrainerState],
    ) -> tensor_dict_type:
        labels = batch[self.labels_key]
        losses_dict = self.loss(forward_results, labels)
        net = forward_results["net"]
        if (
            self.training
            and self._synthetic_step > 0
            and (state is not None and state.step % self._synthetic_step == 0)
        ):
            with timing_context(self, "synthetic.forward"):
                net_min = torch.min(net, dim=0)[0]
                net_max = torch.max(net, dim=0)[0]
                net_diff = net_max - net_min
                diff_span = 0.5 * (self._synthetic_range - 1.0) * net_diff
                synthetic_net = net.new_empty(net.shape)
                synthetic_net.uniform_(0, 1)
                synthetic_net = self._synthetic_range * synthetic_net * net_diff
                synthetic_net = synthetic_net - (diff_span - net_min)
                # synthetic_net ~ U_[ -diff_span + min, diff_span + max ]
                synthetic_outputs = self._core(
                    net.shape[0],
                    synthetic_net,
                    labels,
                    True,
                )
            with timing_context(self, "synthetic.loss"):
                syn_losses_dict = self.loss(  # type: ignore
                    synthetic_outputs,
                    labels,
                    is_synthetic=True,
                )
            losses_dict.update(syn_losses_dict)
            losses_dict["loss"] = losses_dict["loss"] + syn_losses_dict["loss"]
        losses_dict = {k: v.mean() for k, v in losses_dict.items()}
        if not self.training and self.fetch_q:
            q_losses = []
            labels = to_numpy(labels)
            for q in np.linspace(0.05, 0.95, 10):
                q = q.item()
                self.q_metric.config["q"] = q
                pack = self._quantile(net, self._expand(len(net), q), False)
                yq = pack["y_res"] + pack["median"]
                q_losses.append(self.q_metric.metric(labels, to_numpy(yq)))
            quantile_metric = -sum(q_losses) / len(q_losses) * self.q_metric.sign
            ddr_loss = torch.tensor(quantile_metric, dtype=torch.float32)
            losses_dict["ddr"] = ddr_loss
        self.clear_execute_cache()
        return losses_dict


@ModelBase.register("ddr_q")
class Quantile(DDR):
    def _preset_config(self) -> None:
        head_config = self.get_pipe_config("ddr", "head")
        head_config["fetch_cdf"] = False
        head_config["fetch_q"] = True


@ModelBase.register("ddr_cdf")
class CDF(DDR):
    def _preset_config(self) -> None:
        head_config = self.get_pipe_config("ddr", "head")
        head_config["fetch_cdf"] = True
        head_config["fetch_q"] = False


__all__ = ["CDF", "DDR", "Quantile"]
